#!/usr/bin/env python

""" wicd - wireless connection daemon frontend implementation

This module implements a usermode frontend for wicd.  It updates connection
information, provides an (optional) tray icon, and allows for launching of 
the wicd GUI and Wired Profile Chooser.

class TrayIcon() -- Parent class of TrayIconGUI and IconConnectionInfo.
    class TrayConnectionInfo() -- Child class of TrayIcon which provides
        and updates connection status.
    class TrayIconGUI() -- Child class of TrayIcon which implements the tray.
        icon itself.  Parent class of StatusTrayIconGUI and EggTrayIconGUI.
    class StatusTrayIconGUI() -- Implements the tray icon using a 
                                 gtk.StatusIcon.
    class EggTrayIconGUI() -- Implements the tray icon using egg.trayicon.
    def usage() -- Prints usage information.
    def main() -- Runs the wicd frontend main loop.

"""

#
#   Copyright (C) 2007 - 2009 Adam Blackburn
#   Copyright (C) 2007 - 2009 Dan O'Reilly
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License Version 2 as
#   published by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys
import gtk
import gobject
import getopt
import os
import pango
import time
import atexit
from dbus import DBusException

import pygtk
pygtk.require('2.0')

HAS_NOTIFY = True
try:
    import pynotify
    if not pynotify.init("Wicd"):
        HAS_NOTIFY = False
except ImportError:
    HAS_NOTIFY = False

# Wicd specific imports
from wicd import wpath
from wicd import misc
from wicd import dbusmanager
import gui
from guiutil import error, can_use_notify

from wicd.translations import _

ICON_AVAIL = True
USE_EGG = False
# Import egg.trayicon if we're using an older gtk version
if not hasattr(gtk, "StatusIcon"):
    try:
        import egg.trayicon
        USE_EGG = True
    except ImportError:
        print 'Unable to load tray icon: Missing both egg.trayicon and gtk.StatusIcon modules.'
        ICON_AVAIL = False

misc.RenameProcess("wicd-client")

if __name__ == '__main__':
    wpath.chdir(__file__)
    
daemon = wireless = wired = lost_dbus_id = None
DBUS_AVAIL = False

def catchdbus(func):
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except DBusException, e:
            if e.get_dbus_name() != None and "DBus.Error.AccessDenied" in e.get_dbus_name():
                error(None, _('Unable to contact the Wicd daemon due to an access denied error from DBus. Please check that your user is in the $A group.').replace("$A","<b>"+wpath.wicd_group+"</b>"))
                #raise
                raise DBusException(e)
            else:
                print "warning: ignoring exception %s" % e
            return None
    wrapper.__name__ = func.__name__
    wrapper.__module__ = func.__module__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    return wrapper
            

class NetworkMenuItem(gtk.ImageMenuItem):
    def __init__(self, lbl, is_active=False):
        gtk.ImageMenuItem.__init__(self)
        self.label = gtk.Label(lbl)
        if is_active:
            atrlist = pango.AttrList()
            atrlist.insert(pango.AttrWeight(pango.WEIGHT_BOLD, 0, 50))
            self.label.set_attributes(atrlist)
        self.label.set_justify(gtk.JUSTIFY_LEFT)
        self.label.set_alignment(0, 0)
        self.add(self.label)
        self.label.show()
        

class TrayIcon(object):
    """ Base Tray Icon class.
    
    Base Class for implementing a tray icon to display network status.
    
    """
    def __init__(self, animate, displaytray=True, displayapp=False):
        self.cur_sndbytes = -1
        self.cur_rcvbytes = -1
        self.last_sndbytes = -1
        self.last_rcvbytes = -1
        self.max_snd_gain = 10000
        self.max_rcv_gain = 10000

        if USE_EGG:
            self.tr = self.EggTrayIconGUI(self)
        else:
            self.tr = self.StatusTrayIconGUI(self)
        if displayapp:
            self.tr.toggle_wicd_gui()
        self.icon_info = self.TrayConnectionInfo(self, self.tr, animate)
        self.tr.icon_info = self.icon_info
        print 'displaytray %s' % displaytray
        self.tr.visible(displaytray)
        
    def is_embedded(self):
        if USE_EGG:
            raise NotImplementedError()
        else:
            return self.tr.is_embedded()
   
    def get_bandwidth_bytes(self):
        """ Gets the amount of byte sent sine the last time I checked """
        dev_dir = '/sys/class/net/'
        iface = daemon.GetCurrentInterface()

        for fldr in os.listdir(dev_dir):
            if fldr == iface:
                dev_dir = dev_dir + fldr + "/statistics/"
                break
      
        try:
            self.cur_rcvbytes = int(open(dev_dir + "rx_bytes", "r").read().strip())
            self.cur_sndbytes = int(open(dev_dir + "tx_bytes", "r").read().strip())
        except:
           self.cur_sndbytes = -1
           self.cur_rcvbytes = -1

    class TrayConnectionInfo(object):
        """ Class for updating the tray icon status. """
        def __init__(self, parent, tr, animate=True):
            """ Initialize variables needed for the icon status methods. """
            self.last_strength = -2
            self.still_wired = False
            self.network = ''
            self.tried_reconnect = False
            self.connection_lost_counter = 0
            self.tr = tr
            self.last_sndbytes = -1
            self.last_rcvbytes = -1
            self.max_snd_gain = 10000
            self.max_rcv_gain = 10000
            self.animate = animate

            self.parent = parent

            self.network_name = ''          # SSID
            self.network_type = 'none'      # Wired/Wireless/None
            self.network_str = ''           # Signal Strength
            self.network_addr = '0.0.0.0'   # IP Address
            self.network_br = ''            # Bitrate

            # keep track of the last state to provide appropriate
            # notifications
            self._last_bubble = None
            self.last_state = None
            self.should_notify = True

            if DBUS_AVAIL:
                self.update_tray_icon()
            else:
                handle_no_dbus()
                self.set_not_connected_state()

            # Initial update of the tooltip
            self.update_tooltip()

        def update_tooltip(self):
            """
            Updates the trayicon tooltip based on current connection status
            """
            if (self.network_type == "none"):
                self.tr.set_tooltip(_('Not connected'))
            elif (self.network_type == "wireless"):
                self.tr.set_tooltip(_('Connected to $A at $B (IP: $C)')
                        .replace('$A', self.network_name)
                        .replace('$B', self.network_str)
                        .replace('$C', self.network_addr))
            elif (self.network_type == "wired"):
                self.tr.set_tooltip(_('Connected to wired network (IP: $A)')
                        .replace('$A', self.network_addr))
            elif (self.network_type == "killswitch"):
                self.tr.set_tooltip(_('Not connected') + "(" +
                        _('Wireless Kill Switch Enabled') + ")")
            elif (self.network_type == "no_daemon"):
                self.tr.set_tooltip(_('Wicd daemon unreachable'))

            return True

        def _show_notification(self, title, details, image=None):
            if self.should_notify:
                try:
                    if not self._last_bubble:
                        self._last_bubble = pynotify.Notification(title, details,
                                                                  image)
                        self._last_bubble.show()
                    else:
                        self._last_bubble.clear_actions()
                        self._last_bubble.clear_hints()
                        self._last_bubble.update(title, details, image)
                        self._last_bubble.show()
                except Exception, e:
                    if hasattr(e, 'message') and e.message != '':
                        msg = e.message
                    elif hasattr(e, 'args') and len(e.args) > 0:
                        msg = e.args[-1]
                    else:
                        msg = str(e)
                    print "Exception during notification: %s" % msg

                self.should_notify = False

        @catchdbus
        def wired_profile_chooser(self):
            """ Launch the wired profile chooser. """
            gui.WiredProfileChooser()
            daemon.SetNeedWiredProfileChooser(False)
            
        def set_wired_state(self, info):
            """ Sets the icon info for a wired state. """
            wired_ip = info[0]
            self.network_addr = str(info[0])
            self.network_type = "wired"
            self.tr.set_from_file(os.path.join(wpath.images, "wired.png"))
            # status_string = _('Connected to wired network (IP: $A)').replace('$A',
            #wired_ip)
            # self.tr.set_tooltip(status_string)
            self._show_notification(_('Wired Network'),
                                    _('Connection established'),
                                    'network-wired')
            
            self.update_tooltip()

        @catchdbus
        def set_wireless_state(self, info):
            """ Sets the icon info for a wireless state. """
            lock = ''
            wireless_ip = info[0]
            self.network = info[1]
            strength = info[2]
            cur_net_id = int(info[3])
            sig_string = daemon.FormatSignalForPrinting(str(strength))
            self.network_type = "wireless"
            self.network_addr = str(info[0])
            self.network_name = info[1]
            self.network_str = sig_string
            self.network_br = info[4]
            self.set_signal_image(int(info[2]), lock)
            
            if wireless.GetWirelessProperty(cur_net_id, "encryption"):
                lock = "-lock"
            # status_string = (_('Connected to $A at $B (IP: $C)')
            #.replace('$A', self.network)
            #                    .replace('$B', sig_string)
            #                    .replace('$C', str(wireless_ip))) 
            #self.tr.set_tooltip(status_string)
            self.set_signal_image(int(strength), lock)
            self._show_notification(self.network,
                                    _('Connection established'),
                                    'network-wireless')
            
                
            self.update_tooltip()
            
        def set_connecting_state(self, info):
            """ Sets the icon info for a connecting state. """
            wired = False
            if info[0] == 'wired' and len(info) == 1:
                cur_network = _('Wired Network')
                wired = True
            else:
                cur_network = info[1]
            status_string = _('Connecting') + " to " + \
                                cur_network + "..."
            self.update_tooltip()
            # self.tr.set_tooltip(status_string)
            self.tr.set_from_file(os.path.join(wpath.images, "no-signal.png"))
            if wired:
                self._show_notification(cur_network,
                                        _('Establishing connection...'),
                                        'network-wired')
            else:
                self._show_notification(cur_network,
                                        _('Establishing connection...'),
                                        'network-wireless')

            
        @catchdbus
        def set_not_connected_state(self, info=None):
            """ Set the icon info for the not connected state. """
            self.tr.set_from_file(wpath.images + "no-signal.png")
            if not DBUS_AVAIL:
                status = _('Wicd daemon unreachable')
            elif wireless.GetKillSwitchEnabled():
                status = (_('Not connected') + " (" + 
                         _('Wireless Kill Switch Enabled') + ")")
            else:
                status = _('Not connected')
            # self.tr.set_tooltip(status)
            self._show_notification(_('Disconnected'), None, 'stop')
            self.update_tooltip()

        @catchdbus
        def update_tray_icon(self, state=None, info=None):
            """ Updates the tray icon and current connection status. """
            if not DBUS_AVAIL: return False

            if not state or not info:
                [state, info] = daemon.GetConnectionStatus()

            # should this state change display a notification?
            self.should_notify = (can_use_notify() and 
                                  self.last_state != state)

            self.last_state = state
            
            if state == misc.WIRED:
                self.set_wired_state(info)
            elif state == misc.WIRELESS:
                self.set_wireless_state(info)
            elif state == misc.CONNECTING:
                self.set_connecting_state(info)
            elif state in (misc.SUSPENDED, misc.NOT_CONNECTED):
                self.set_not_connected_state(info)
            else:
                print 'Invalid state returned!!!'
                return False
            return True

        @catchdbus
        def set_signal_image(self, wireless_signal, lock):
            """ Sets the tray icon image for an active wireless connection. """
            if self.animate:
                TrayIcon.get_bandwidth_bytes(self.parent)
                prefix = self.get_bandwidth_activity()
            else:
                prefix = 'idle-'
            if daemon.GetSignalDisplayType() == 0:
                if wireless_signal > 75:
                    signal_img = "high-signal"
                elif wireless_signal > 50:
                    signal_img = "good-signal"
                elif wireless_signal > 25:
                    signal_img = "low-signal"
                else:
                    signal_img = "bad-signal"
            else:
                if wireless_signal >= -60:
                    signal_img = "high-signal"
                elif wireless_signal >= -70:
                    signal_img = "good-signal"
                elif wireless_signal >= -80:
                    signal_img = "low-signal"
                else:
                    signal_img = "bad-signal"

            img_file = ''.join([wpath.images, prefix, signal_img, lock, ".png"])
            self.tr.set_from_file(img_file)
        
        @catchdbus
        def get_bandwidth_activity(self):
            """ Determines what network activity state we are in. """
            transmitting = False
            receiving = False
                    
            dev_dir = '/sys/class/net/'
            wiface = daemon.GetWirelessInterface()
            for fldr in os.listdir(dev_dir):
                if fldr == wiface:
                    dev_dir = dev_dir + fldr + "/statistics/"
                    break
            try:
                rcvbytes = int(open(dev_dir + "rx_bytes", "r").read().strip())
                sndbytes = int(open(dev_dir + "tx_bytes", "r").read().strip())
            except IOError:
                sndbytes = None
                rcvbytes = None
                
            if not rcvbytes or not sndbytes:
                return 'idle-'
                    
            # Figure out receiving data info.
            activity = self.is_network_active(self.parent.cur_rcvbytes, 
                                              self.parent.max_rcv_gain,
                                              self.parent.last_rcvbytes)
            receiving = activity[0]
            self.parent.max_rcv_gain = activity[1]
            self.parent.last_rcvbytes = activity[2]
                    
            # Figure out out transmitting data info.
            activity = self.is_network_active(self.parent.cur_sndbytes, 
                                              self.parent.max_snd_gain,
                                              self.parent.last_sndbytes)
            transmitting = activity[0]
            self.parent.max_snd_gain = activity[1]
            self.parent.last_sndbytes = activity[2]
                    
            if transmitting and receiving:
                return 'both-'
            elif transmitting:
                return 'transmitting-'
            elif receiving:
                return 'receiving-'
            else:
                return 'idle-'
            
        def is_network_active(self, bytes, max_gain, last_bytes):
            """ Determines if a network is active.
            
            Determines if a network is active by looking at the
            number of bytes sent since the previous check.  This method
            is generic, and can be used to determine activity in both
            the sending and receiving directions.
            
            Returns:
            A tuple containing three elements:
            1) a boolean specifying if the network is active.
            2) an int specifying the maximum gain the network has had.
            3) an int specifying the last recorded number of bytes sent.
            
            """
            active = False
            if last_bytes == -1:
                last_bytes = bytes
            elif bytes > (last_bytes + float(max_gain / 20.0)):
                last_bytes = bytes
                active = True
                
                gain = bytes - last_bytes
                if gain > max_gain:
                    max_gain = gain
            return (active, max_gain, last_bytes)


    class TrayIconGUI(object):
        """ Base Tray Icon UI class.
        
        Implements methods and variables used by both egg/StatusIcon
        tray icons.

        """
        def __init__(self, parent):
            menu = """
                    <ui>
                        <menubar name="Menubar">
                            <menu action="Menu">
                                <menu action="Connect">
                                </menu>
                                <separator/>
                                <menuitem action="Info"/>
                                <menuitem action="Quit"/>
                            </menu>
                        </menubar>
                    </ui>
            """
            actions = [
                    ('Menu',  None, 'Menu'),
                    ('Connect', gtk.STOCK_CONNECT, _('Connect')),
                    ('Info', gtk.STOCK_INFO, _('_Connection Info'), None,
                     _('Information about the current connection'),
                     self.on_conn_info),
                    ('Quit',gtk.STOCK_QUIT,_('_Quit'),None,_('Quit wicd-tray-icon'),
                     self.on_quit),
                    ]
            actg = gtk.ActionGroup('Actions')
            actg.add_actions(actions)
            self.manager = gtk.UIManager()
            self.manager.insert_action_group(actg, 0)
            self.manager.add_ui_from_string(menu)
            self.menu = (self.manager.get_widget('/Menubar/Menu/Quit').
                                                                  props.parent)
            self.gui_win = None
            self.current_icon_path = None
            self._is_scanning = False
            net_menuitem = self.manager.get_widget("/Menubar/Menu/Connect/")
            net_menuitem.connect("activate", self.on_net_menu_activate)
            
            self.parent = parent
            self.time = 2           # Time between updates
            self.cont = 'Stop'
            self.conn_info_txt = ''
            
        def tray_scan_started(self):
            """ Callback for when a wireless scan is started. """
            if not DBUS_AVAIL: return
            self._is_scanning = True
            self.init_network_menu()
            
        def tray_scan_ended(self):
            """ Callback for when a wireless scan finishes. """
            if not DBUS_AVAIL: return
            self._is_scanning = False
            self.populate_network_menu()
                
        def on_activate(self, data=None):
            """ Opens the wicd GUI. """
            if DBUS_AVAIL:
                self.toggle_wicd_gui()
            else:
                # error(None, _('The wicd daemon is unavailable, so your request cannot be completed'))
                pass

        def on_quit(self, widget=None):
            """ Closes the tray icon. """
            sys.exit(0)

        def on_about(self, data=None):
            """ Opens the About Dialog. """
            dialog = gtk.AboutDialog()
            dialog.set_name('Wicd Tray Icon')
            dialog.set_version('2.0')
            dialog.set_comments('An icon that shows your network connectivity')
            dialog.set_website('http://wicd.net')
            dialog.run()
            dialog.destroy()

        def on_conn_info(self, data=None):
            """ Opens the Connection Information Dialog """
            window = gtk.Dialog("Wicd Connection Info", None, 0, (gtk.STOCK_OK, gtk.RESPONSE_CLOSE))
           
            # Create labels
            self.label = gtk.Label()
            self.data = gtk.Label()
            self.data.set_selectable(True)
            self.label.show()
            self.data.show()
            self.list = []
            self.list.append(self.data)
            self.list.append(self.label)

            # Setup table
            table = gtk.Table(1,2)
            table.set_col_spacings(12)
            table.attach(self.label, 0, 1, 0, 1)
            table.attach(self.data, 1, 2, 0 ,1)

            # Setup Window
            content = window.get_content_area()
            content.pack_start(table, True, True, 0)
            content.show_all()
            
            # Start updates
            self.cont = 'Go'
            gobject.timeout_add(5000, self.update_conn_info_win, self.list)
            self.update_conn_info_win(self.list)
            
            window.run()
    
            # Destroy window and stop updates
            window.destroy()
            self.cont = 'Stop'
    
        def update_conn_info_win(self, list): 
            """ Updates the information in the connection summary window """
            if (self.cont == "Stop"):
                return False
            
            [state, info] = daemon.GetConnectionStatus()
            [rx, tx] = self.get_current_bandwidth()
                   
            # Choose info for the data
            if state == misc.WIRED:
                text = (_('''$A
$B KB/s
$C KB/s''')
                        .replace('$A', str(info[0]))    #IP
                        .replace('$B', str(rx))         #RX
                        .replace('$C', str(tx)))        #TX
            elif state == misc.WIRELESS:
                text = (_('''$A
$B
$C
$D
$E KB/s
$F KB/s''')
                        .replace('$A', str(info[1]))    #SSID
                        .replace('$B', str(info[4]))    #Speed
                        .replace('$C', str(info[0]))    #IP
                        .replace('$D', daemon.FormatSignalForPrinting(str(info[2])))
                        .replace('$E', str(rx))
                        .replace('$F', str(tx)))
            else:
                text = ''

            # Choose info for the labels
            self.list[0].set_text('\n' + text)
            if state == misc.WIRED:
                self.list[1].set_text(_('''Wired
IP:
RX:
TX:'''))
            elif state == misc.WIRELESS:
                self.list[1].set_text(_('''Wireless
SSID:
Speed:
IP:
Strength:
RX:
TX:'''))
            elif state == misc.CONNECTING:
                self.list[1].set_text(_('Connecting'))
            elif state in (misc.SUSPENDED, misc.NOT_CONNECTED):
                self.list[1].set_text(_('Disconnected'))
                       
            return True 
                            
        def get_current_bandwidth(self):
            """ 
            Calculates the current bandwidth based on sent/received bytes
            divided over time. Unit is in KB/s
            """
            self.parent.get_bandwidth_bytes()
            rxb = self.parent.cur_rcvbytes - self.parent.last_rcvbytes
            txb = self.parent.cur_sndbytes - self.parent.last_sndbytes

            self.parent.last_rcvbytes = self.parent.cur_rcvbytes
            self.parent.last_sndbytes = self.parent.cur_sndbytes

            rx_rate = float(rxb / (self.time * 1024))
            tx_rate = float(txb / (self.time * 1024))
                                
            return (rx_rate, tx_rate)
        
        def _add_item_to_menu(self, net_menu, lbl, type_, n_id, is_connecting, 
                              is_active):
            """ Add an item to the network list submenu. """
            def network_selected(widget, net_type, net_id):
                """ Callback method for a menu item selection. """
                if net_type == "__wired__":
                    wired.ConnectWired()
                else:
                    wireless.ConnectWireless(net_id)
                    
            item = NetworkMenuItem(lbl, is_active)
            image = gtk.Image()
            
            if type_ == "__wired__":
                image.set_from_icon_name("network-wired", 2)
            else:
                pb = gtk.gdk.pixbuf_new_from_file_at_size(self._get_img(n_id),
                                                          20, 20)
                image.set_from_pixbuf(pb)
                del pb
            item.set_image(image)
            del image
            item.connect("activate", network_selected, type_, n_id)
            net_menu.append(item)
            item.show()
            if is_connecting:
                item.set_sensitive(False)  
            del item
                
        @catchdbus
        def _get_img(self, net_id):
            """ Determines which image to use for the wireless entries. """
            def fix_strength(val, default):
                """ Assigns given strength to a default value if needed. """
                return val and int(val) or default
            
            def get_prop(prop):
                return wireless.GetWirelessProperty(net_id, prop)
            
            strength = fix_strength(get_prop("quality"), -1)
            dbm_strength = fix_strength(get_prop('strength'), -100)

            if daemon.GetWPADriver() == 'ralink legacy' or \
               daemon.GetSignalDisplayType() == 1:
                if dbm_strength >= -60:
                    signal_img = 'signal-100.png'
                elif dbm_strength >= -70:
                    signal_img = 'signal-75.png'
                elif dbm_strength >= -80:
                    signal_img = 'signal-50.png'
                else:
                    signal_img = 'signal-25.png'
            else:
                if strength > 75:
                    signal_img = 'signal-100.png'
                elif strength > 50:
                    signal_img = 'signal-75.png'
                elif strength > 25:
                    signal_img = 'signal-50.png'
                else:
                    signal_img = 'signal-25.png'
            return wpath.images + signal_img
            
        @catchdbus
        def on_net_menu_activate(self, item):
            """ Trigger a background scan to populate the network menu. 
            
            Called when the network submenu is moused over.  We
            sleep briefly, clear pending gtk events, and if
            we're still being moused over we trigger a scan.
            This is to prevent scans when the user is just
            mousing past the menu to select another menu item.
            
            """
            def dummy(x=None): pass
            
            if self._is_scanning:
                return True
            
            self.init_network_menu()
            gobject.timeout_add(800, self._trigger_scan_if_needed, item)
            
        @catchdbus
        def _trigger_scan_if_needed(self, item):
            """ Trigger a scan if the network menu is being hovered over. """
            while gtk.events_pending():
                gtk.main_iteration()
            if item.state != gtk.STATE_PRELIGHT:
                return True
            wireless.Scan(False)
            return False
        
        @catchdbus
        def populate_network_menu(self, data=None):
            """ Populates the network list submenu. """
            def get_prop(net_id, prop):
                return wireless.GetWirelessProperty(net_id, prop)

            net_menuitem = self.manager.get_widget("/Menubar/Menu/Connect/")
            submenu = net_menuitem.get_submenu()
            self._clear_menu(submenu)
            if not DBUS_AVAIL:
                net_menuitem.show()
                return

            is_connecting = daemon.CheckIfConnecting()
            num_networks = wireless.GetNumberOfNetworks()
            [status, info] = daemon.GetConnectionStatus()
                
            if daemon.GetAlwaysShowWiredInterface() or \
               wired.CheckPluggedIn():
                if status == misc.WIRED:
                    is_active = True
                else:
                    is_active = False
                self._add_item_to_menu(submenu, "Wired Network", "__wired__", 0,
                                       is_connecting, is_active)
                sep = gtk.SeparatorMenuItem()
                submenu.append(sep)
                sep.show()
            
            if num_networks > 0:
                skip_never_connect = not daemon.GetShowNeverConnect()
                for x in xrange(0, num_networks):
                    if skip_never_connect and misc.to_bool(get_prop(x,"never")): continue
                    essid = get_prop(x, "essid")
                    if status == misc.WIRELESS and info[1] == essid:
                        is_active = True
                    else:
                        is_active = False
                    self._add_item_to_menu(submenu, essid, "wifi", x,
                                           is_connecting, is_active)
            else:
                no_nets_item = gtk.MenuItem(_('No wireless networks found.'))
                no_nets_item.set_sensitive(False)
                no_nets_item.show()
                submenu.append(no_nets_item)
                    
            net_menuitem.show()
        
        def init_network_menu(self):
            """ Set the right-click network menu to the scanning state. """
            net_menuitem = self.manager.get_widget("/Menubar/Menu/Connect/")
            submenu = net_menuitem.get_submenu()
            self._clear_menu(submenu)

            loading_item = gtk.MenuItem(_('Scanning') + "...")
            loading_item.set_sensitive(False)
            loading_item.show()
            submenu.append(loading_item)
            net_menuitem.show()
            
        def _clear_menu(self, menu):
            """ Clear the right-click menu. """
            for item in menu.get_children():
                menu.remove(item)
                item.destroy()
            
        def toggle_wicd_gui(self):
            """ Toggles the wicd GUI. """
            if not self.gui_win:
                self.gui_win = gui.appGui(tray=self)
            elif not self.gui_win.is_visible:
                self.gui_win.show_win()
            else:
                self.gui_win.exit()
                return True
    

    if USE_EGG:
        class EggTrayIconGUI(TrayIconGUI):
            """ Tray Icon for gtk < 2.10.
            
            Uses the deprecated egg.trayicon module to implement the tray icon.
            Since it relies on a deprecated module, this class is only used
            for machines running versions of GTK < 2.10.
            
            """
            def __init__(self, parent):
                """Initializes the tray icon"""
                TrayIcon.TrayIconGUI.__init__(self, parent)
                self.tooltip = gtk.Tooltips()
                self.eb = gtk.EventBox()
                self.tray = egg.trayicon.TrayIcon("WicdTrayIcon")
                self.pic = gtk.Image()
                self.tooltip.set_tip(self.eb, "Initializing wicd...")
                self.pic.set_from_file(wpath.images + "no-signal.png")

                self.eb.connect('button_press_event', self.tray_clicked)
                self.eb.add(self.pic)
                self.tray.add(self.eb)
                self.tray.show_all()

            def tray_clicked(self, widget, event):
                """ Handles tray mouse click events. """
                if event.button == 1:
                    self.toggle_wicd_gui()
                elif event.button == 3:
                    self.init_network_menu()
                    self.menu.popup(None, None, None, event.button, event.time)

            def set_from_file(self, val=None):
                """ Calls set_from_file on the gtk.Image for the tray icon. """
                self.pic.set_from_file(val)

            def set_tooltip(self, val):
                """ Set the tooltip for this tray icon.
                
                Sets the tooltip for the gtk.ToolTips associated with this
                tray icon.

                """
                self.tooltip.set_tip(self.eb, val)

            def visible(self, val):
                """ Set if the icon is visible or not.

                If val is True, makes the icon visible, if val is False,
                hides the tray icon.

                """
                if val:
                    self.tray.show_all()
                else:
                    self.tray.hide_all()


    if hasattr(gtk, "StatusIcon"):
        class StatusTrayIconGUI(gtk.StatusIcon, TrayIconGUI):
            """ Class for creating the wicd tray icon on gtk > 2.10.
            
            Uses gtk.StatusIcon to implement a tray icon.
            
            """
            def __init__(self, parent):
                TrayIcon.TrayIconGUI.__init__(self, parent)
                gtk.StatusIcon.__init__(self)

                self.current_icon_path = ''
                self.set_visible(True)
                self.connect('activate', self.on_activate)
                self.connect('popup-menu', self.on_popup_menu)
                self.set_from_file(wpath.images + "no-signal.png")
                self.set_tooltip("Initializing wicd...")

            def on_popup_menu(self, status, button, timestamp):
                """ Opens the right click menu for the tray icon. """
                self.init_network_menu()
                self.menu.popup(None, None, None, button, timestamp)

            def set_from_file(self, path=None):
                """ Sets a new tray icon picture. """
                if path != self.current_icon_path:
                    self.current_icon_path = path
                    gtk.StatusIcon.set_from_file(self, path)

            def visible(self, val):
                """ Set if the icon is visible or not.

                If val is True, makes the icon visible, if val is False,
                hides the tray icon.

                """
                self.set_visible(val)

def usage():
    """ Print usage information. """
    print """
wicd %s 
wireless (and wired) connection daemon front-end.

Arguments:
\t-t\t--tray\tRun the wicd tray icon only.
\t-n\t--no-tray\tRun wicd without the tray icon.
\t-h\t--help\t\tPrint this help information.
\t-a\t--no-animate\tRun the tray without network traffic tray animations.
\t-o\t--only-notifications\tDon't display anything except notifications.
""" % wpath.version

def setup_dbus(force=True):
    global daemon, wireless, wired, DBUS_AVAIL, lost_dbus_id
    print "Connecting to daemon..."
    try:
        dbusmanager.connect_to_dbus()
    except DBusException:
        if force:
            print "Can't connect to the daemon, trying to start it automatically..."
            misc.PromptToStartDaemon()
            try:
                dbusmanager.connect_to_dbus()
            except DBusException:
                error(None, _("Could not connect to wicd's D-Bus interface. Check the wicd log for error messages."))
                return False
        else:  
            return False
                
    if lost_dbus_id:
        gobject.source_remove(lost_dbus_id)
        lost_dbus_id = None
    dbus_ifaces = dbusmanager.get_dbus_ifaces()
    daemon = dbus_ifaces['daemon']
    wireless = dbus_ifaces['wireless']
    wired = dbus_ifaces['wired']
    DBUS_AVAIL = True
    print "Connected."
    return True

def on_exit():
    if DBUS_AVAIL:
        try:
            daemon.SetGUIOpen(False)
        except DBusException:
            pass

def handle_no_dbus():
    """ Called when dbus announces its shutting down. """
    global DBUS_AVAIL, lost_dbus_id
    DBUS_AVAIL = False
    gui.handle_no_dbus(from_tray=True)
    print "Wicd daemon is shutting down!"
    lost_dbus_id = misc.timeout_add(5, lambda:error(None, _('The wicd daemon has shut down. The UI will not function properly until it is restarted.'),
                                                    block=False))
    return False

@catchdbus
def main(argv):
    """ The main frontend program.

    Keyword arguments:
    argv -- The arguments passed to the script.

    """
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'tnhao', ['help', 'no-tray',
                                                          'tray',
                                                          'no-animate',
                                                          'only-notifications'])
    except getopt.GetoptError:
        # Print help information and exit
        usage()
        sys.exit(2)

    use_tray = True
    animate = True
    display_app = True
    for opt, a in opts:
        if opt in ('-h', '--help'):
            usage()
            sys.exit(0)
        elif opt in ('-t', '--tray'):
            display_app = False
        elif opt in ('-n', '--no-tray'):
            use_tray = False
        elif opt in ('-a', '--no-animate'):
            animate = False
        elif opt in ('-o', '--only-notifications'):
            print "only displaying notifications"
            use_tray = False
            display_app = False
        else:
            usage()
            sys.exit(2)
    
    print 'Loading...'
    setup_dbus()
    atexit.register(on_exit)

    if display_app and not use_tray or not ICON_AVAIL:
        the_gui = gui.appGui(standalone=True)
        mainloop = gobject.MainLoop()
        mainloop.run()
        sys.exit(0)

    # Set up the tray icon GUI and backend
    tray_icon = TrayIcon(animate, displaytray=use_tray, displayapp=display_app)

    # Check to see if wired profile chooser was called before icon
    # was launched (typically happens on startup or daemon restart).
    if DBUS_AVAIL and daemon.GetNeedWiredProfileChooser():
        daemon.SetNeedWiredProfileChooser(False)
        tray_icon.icon_info.wired_profile_chooser()
        
    bus = dbusmanager.get_bus()
    bus.add_signal_receiver(tray_icon.icon_info.wired_profile_chooser,
                            'LaunchChooser', 'org.wicd.daemon')
    bus.add_signal_receiver(tray_icon.icon_info.update_tray_icon,
                            'StatusChanged', 'org.wicd.daemon')
    bus.add_signal_receiver(tray_icon.tr.tray_scan_ended, 'SendEndScanSignal',
                            'org.wicd.daemon.wireless')
    bus.add_signal_receiver(tray_icon.tr.tray_scan_started,
                            'SendStartScanSignal', 'org.wicd.daemon.wireless')
    bus.add_signal_receiver(lambda: (handle_no_dbus() or 
                                     tray_icon.icon_info.set_not_connected_state()), 
                            "DaemonClosing", 'org.wicd.daemon')
    bus.add_signal_receiver(lambda: setup_dbus(force=False), "DaemonStarting",
                            "org.wicd.daemon")
    print 'Done loading.'
    mainloop = gobject.MainLoop()
    mainloop.run()


if __name__ == '__main__':
    main(sys.argv)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" networking - Provides wrappers for common network operations

This module provides wrappers of the common network tasks as well as
threads to perform the actual connecting to networks.

class Controller() -- Parent class to Wireless and Wired
class ConnectThread() -- Parent class to WirelessConnectThread and
    WiredConnectThread
class Wireless() -- Wrapper for various wireless functions
class Wired() -- Wrapper for various wired functions
class WirelessConnectThread() -- Connection thread for wireless
    interface
class WiredConnectThread() -- Connection thread for wired
    interface

"""

#
#   Copyright (C) 2007 - 2009 Adam Blackburn
#   Copyright (C) 2007 - 2009 Dan O'Reilly
#   Copyright (C) 2007 - 2009 Byron Hillis
#   Copyright (C)        2009 Andrew Psaltis
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License Version 2 as
#   published by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Much thanks to wieman01 for help and support with various types of encyption.
# Also thanks to foxy123, yopnono, and the many others who reported bugs helped
# and helped keep this project moving.
#

import re
import time
import threading
import os
from signal import SIGTERM

# wicd imports 
import misc
import wpath
from backend import BackendManager
from translations import _

if __name__ == '__main__':
    wpath.chdir(__file__)

BACKEND = None
BACKEND_MGR = BackendManager()

def abortable(func):
    """ Mark a method in a ConnectionThread as abortable. 
    
    This decorator runs a check that will abort the connection thread
    if necessary before running a given method.
    
    """
    def wrapper(self, *__args, **__kargs):
        self.abort_if_needed()
        return func(self, *__args, **__kargs)
    
    wrapper.__name__ = func.__name__
    wrapper.__dict__ = func.__dict__
    wrapper.__doc__ = func.__doc__
    wrapper.__module = func.__module__
    return wrapper

def get_backend_list():
    """ Returns a list of available backends. """
    if BACKEND_MGR:
        return BACKEND_MGR.get_available_backends()
    else:
        return [""]
    
def get_backend_update_interval():
    """ Returns the suggested connection status update interval. """
    if BACKEND_MGR:
        return BACKEND_MGR.get_update_interval()
    else:
        return 5  # Seconds, this should never happen though.
    
def get_current_backend():
    """ Returns the current backend instance. """
    if BACKEND_MGR:
        return BACKEND_MGR.get_current_backend()
    else:
        return None
    
def get_backend_description(backend_name):
    """ Returns the description of the currently loaded backend. """
    return BACKEND_MGR.get_backend_description(backend_name)

def get_backend_description_dict():
    """ Returns a dict of all available backend descriptions. """
    d = {}
    for be in get_backend_list():
        if be:
            d[be] = get_backend_description(be)
    return d

def expand_script_macros(script, msg, bssid, essid):
    """Expands any supported macros in a script.

    Keyword arguments:
    script -- the script to execute.
    msg -- the name of the script, %{script} will be expanded to this.
    bssid -- the bssid of the network we connect to, defaults to 'wired'.
    essid -- the essid of the network we connect to, defaults to 'wired'.
    
    """
    def repl(match):
        macro = match.group(1).lower()
        if macro_dict.has_key(macro):
            return macro_dict[macro]
        print 'Warning: found illegal macro %s in %s script' % (macro, msg)
        return match.group()
    
    macro_dict = { 'script' : msg,
             'bssid' : bssid,
             'essid' : essid }
    regex = re.compile(r'%\{([a-zA-Z0-9]+)\}')
    expanded = regex.sub(repl, script)
    print "Expanded '%s' to '%s'" % (script, expanded)
    return expanded

 
class Controller(object):
    """ Parent class for the different interface types. """
    def __init__(self, debug=False):
        """ Initialise the class. """
        self.global_dns_1 = None
        self.global_dns_2 = None
        self.global_dns_3 = None
        self.global_dns_dom = None
        self.global_search_dom = None
        self._dhcp_client = None
        self._flush_tool = None
        self._debug = debug
        self._backend = None
        self.connecting_thread = None
        self.before_script = None
        self.after_script = None
        self.pre_disconnect_script = None
        self.post_disconnect_script = None
        self.driver = None
        self.iface = None
    
    def get_debug(self): return self._debug
    def set_debug(self, value):
        self._debug = value
        if self.iface:
            self.iface.SetDebugMode(value)
    debug = property(get_debug, set_debug)
    
    def set_dhcp_client(self, value):
        self._dhcp_client = value
        if self.iface:
            self.iface.DHCP_CLIENT = value
    def get_dhcp_client(self): return self._dhcp_client
    dhcp_client = property(get_dhcp_client, set_dhcp_client)
    
    def set_flush_tool(self, value):
        self._flush_tool = value
        if self.iface:
            self.iface.flush_tool = value
    def get_flush_tool(self): return self._flush_tool
    flush_tool = property(get_flush_tool, set_flush_tool)
    
    def LoadBackend(self, backend_name):
        """ Load the given networking backend. """
        global BACKEND
        if backend_name == self._backend:
            return
        self._backend = BACKEND_MGR.load_backend(backend_name)
        BACKEND = self._backend
        
    def NeedsExternalCalls(self):
        """ Returns true if the loaded backend needs external calls. """
        if self._backend:
            return self._backend.NeedsExternalCalls()
        else:
            return True
        
    def GetIP(self, ifconfig=""):
        """ Get the IP of the interface.

        Returns:
        The IP address of the interface in dotted notation.

        """
        return self.iface.GetIP(ifconfig)

    def Disconnect(self, nettype, name, mac):
        """ Disconnect from the network. """
        iface = self.iface
        # mac and name need to be strings
        if mac in (None, ''):
            mac = 'X'
        if name in (None, ''):
            name = 'X'
        misc.ExecuteScripts(wpath.predisconnectscripts, self.debug,
                           extra_parameters=(nettype, name, mac))
        if self.pre_disconnect_script:
            print 'Running pre-disconnect script'
            misc.ExecuteScript(expand_script_macros(self.pre_disconnect_script,
                                                    'pre-disconnection',
                                                    mac, name),
                               self.debug)
        iface.ReleaseDHCP()
        iface.SetAddress('0.0.0.0')
        iface.FlushRoutes()
        iface.Down()
        iface.Up()
        misc.ExecuteScripts(wpath.postdisconnectscripts, self.debug,
                            extra_parameters=(nettype, name, mac))
        if self.post_disconnect_script:
            print 'Running post-disconnect script'
            misc.ExecuteScript(expand_script_macros(self.post_disconnect_script,
                                                    'post-disconnection',
                                                   mac, name),
                               self.debug)
        
    def ReleaseDHCP(self):
        """ Release the DHCP lease for this interface. """
        return self.iface.ReleaseDHCP()
    
    def KillDHCP(self):
        """ Kill the managed DHCP client if its in a connecting state. """
        print 'running kill dhcp.'
        if (self.connecting_thread.is_connecting and 
            self.iface.dhcp_object):
            if self.iface.dhcp_object.poll() is None:
                os.kill(self.iface.dhcp_object.pid, SIGTERM)
                self.iface.dhcp_object = None
    
    def IsUp(self):
        """ Calls the IsUp method for the wired interface.
        
        Returns:
        True if the interface is up, False otherwise.
        
        """
        return self.iface.IsUp()
    
    def EnableInterface(self):
        """ Puts the interface up.
        
        Returns:
        True if the interface was put up succesfully, False otherwise.
        
        """
        return self.iface.Up()
    
    def DisableInterface(self):
        """ Puts the interface down.
        
        Returns:
        True if the interface was put down succesfully, False otherwise.
        
        """
        return self.iface.Down()
    
    def AppAvailable(self, app):
        """ Determine if the given application is installed. """
        return self.iface.AppAvailable(app)
    

class ConnectThread(threading.Thread):
    """ A class to perform network connections in a multi-threaded way.

    Useless on it's own, this class provides the generic functions
    necessary for connecting using a separate thread.
    
    """

    is_connecting = None
    should_die = False
    lock = threading.Lock()

    def __init__(self, network, interface_name, before_script, after_script, 
                 pre_disconnect_script, post_disconnect_script, gdns1,
                 gdns2, gdns3, gdns_dom, gsearch_dom, iface,
                 debug):
        """ Initialise the required object variables and the thread.

        Keyword arguments:
        network -- the network to connect to
        wireless -- name of the wireless interface
        wired -- name of the wired interface
        before_script -- script to run before bringing up the interface
        after_script -- script to run after bringing up the interface
        pre_disconnect_script -- script to run before disconnection
        post_disconnect_script -- script to run after disconnection
        gdns1 -- global DNS server 1
        gdns2 -- global DNS server 2
        gdns3 -- global DNS server 3
        debug -- debug mode status

        """
        threading.Thread.__init__(self)
        self.network = network
        self.is_connecting = False
        self.is_aborted = False
        self.connect_result = None
        self.before_script = before_script
        self.after_script = after_script
        self.pre_disconnect_script = pre_disconnect_script
        self.post_disconnect_script = post_disconnect_script
        self._should_die = False
        self.abort_reason = ""
        self.connect_result = ""

        self.global_dns_1 = gdns1
        self.global_dns_2 = gdns2
        self.global_dns_3 = gdns3
        self.global_dns_dom = gdns_dom
        self.global_search_dom = gsearch_dom

        self.iface = iface

        self.connecting_status = None
        self.debug = debug
        
        self.SetStatus('interface_down')
        
    def run(self):
        self.connect_result = "failed"
        try:
            self._connect()
        finally:
            self.is_connecting = False
        
    def set_should_die(self, val):
        self.lock.acquire()
        try:
            self._should_die = val
        finally:
            self.lock.release()
    def get_should_die(self): return self._should_die
    should_die = property(get_should_die, set_should_die)

    def SetStatus(self, status):
        """ Set the threads current status message in a thread-safe way.

        Keyword arguments:
        status -- the current connection status

        """
        self.lock.acquire()
        try:
            self.connecting_status = status
        finally:
            self.lock.release()

    def GetStatus(self):
        """ Get the threads current status message in a thread-safe way.

        Returns:
        The current connection status.

        """
        self.lock.acquire()
        try:
            status = self.connecting_status
        finally:
            self.lock.release()
        return status
    
    @abortable
    def reset_ip_addresses(self, iface):
        """ Resets the IP addresses for both wired/wireless interfaces.
        
        Sets a false ip so that when we set the real one, the correct
        routing entry is created.
        
        """
        print 'Setting false IP...'
        self.SetStatus('resetting_ip_address')
        iface.SetAddress('0.0.0.0')
    
    @abortable
    def put_iface_down(self, iface):
        """ Puts the given interface down. """
        print 'Putting interface down'
        self.SetStatus('interface_down')
        iface.Down()
        
    @abortable
    def run_global_scripts_if_needed(self, script_dir, extra_parameters=()):
        misc.ExecuteScripts(script_dir, verbose=self.debug,
                            extra_parameters=extra_parameters)

    @abortable
    def run_script_if_needed(self, script, msg, bssid='wired', essid='wired'):
        """ Execute a given script if needed.
        
        Keyword arguments:
        script -- the script to execute, or None/'' if there isn't one.
        msg -- the name of the script to display in the log.
        
        """
        if script:
            print 'Executing %s script' % (msg)
            misc.ExecuteScript(expand_script_macros(script, msg, bssid, essid),
                               self.debug)
        
    @abortable
    def flush_routes(self, iface):
        """ Flush the routes for both wired/wireless interfaces. """
        self.SetStatus('flushing_routing_table')
        print 'Flushing the routing table...'
        iface.FlushRoutes()
        
    @abortable
    def set_broadcast_address(self, iface):
        """ Set the broadcast address for the given interface. """
        if not self.network.get('broadcast') == None:
            self.SetStatus('setting_broadcast_address')
            print 'Setting the broadcast address...' + self.network['broadcast']
            iface.SetAddress(broadcast=self.network['broadcast'])
        
    @abortable
    def set_ip_address(self, iface):
        """ Set the IP address for the given interface. 
        
        Assigns a static IP if one is requested, otherwise calls DHCP.
        
        """
        if self.network.get('ip'):
            self.SetStatus('setting_static_ip')
            print 'Setting static IP : ' + self.network['ip']
            iface.SetAddress(self.network['ip'], self.network['netmask'])
            if self.network.get('gateway'):
                print 'Setting default gateway : ' + self.network['gateway']
                iface.SetDefaultRoute(self.network['gateway'])
        else:
            # Run dhcp...
            self.SetStatus('running_dhcp')
            if self.network.get('usedhcphostname') == None:
                self.network['usedhcphostname'] = False
            if self.network.get('dhcphostname') == None:
                self.network['dhcphostname'] = os.uname()[1]
            if not self.network['usedhcphostname']:
                hname = os.uname()[1]
            else:
                hname = self.network['dhcphostname']
            print "Running DHCP with hostname",hname
            dhcp_status = iface.StartDHCP(hname)
            if dhcp_status in ['no_dhcp_offers', 'dhcp_failed']:
                if self.connect_result != "aborted":
                    self.abort_connection(dhcp_status)
                return

    @abortable
    def set_dns_addresses(self, iface):
        """ Set the DNS address(es).

        If static DNS servers or global DNS servers are specified, set them.
        Otherwise do nothing.
        
        """
        if self.network.get('use_global_dns'):
            iface.SetDNS(misc.Noneify(self.global_dns_1),
                         misc.Noneify(self.global_dns_2), 
                         misc.Noneify(self.global_dns_3),
                         misc.Noneify(self.global_dns_dom),
                         misc.Noneify(self.global_search_dom))
        elif self.network.get('use_static_dns') and (self.network.get('dns1') or
                    self.network.get('dns2') or self.network.get('dns3')):
            self.SetStatus('setting_static_dns')
            iface.SetDNS(self.network.get('dns1'),
                         self.network.get('dns2'),
                         self.network.get('dns3'),
                         self.network.get('dns_domain'),
                         self.network.get('search_domain'))

    @abortable
    def release_dhcp_clients(self, iface):
        """ Release all running dhcp clients. """
        print "Releasing DHCP leases..."
        iface.ReleaseDHCP()
        
    def connect_aborted(self, reason):
        """ Sets the thread status to aborted. """
        if self.abort_reason:
            reason = self.abort_reason
        self.connecting_status = reason
        self.is_aborted = True
        self.connect_result = reason
        self.is_connecting = False
        print 'exiting connection thread'
        
    def abort_connection(self, reason=""):
        """ Schedule a connection abortion for the given reason. """
        self.abort_reason = reason
        self.should_die = True
        
    def abort_if_needed(self):
        """ Abort the thread is it has been requested. """
        self.lock.acquire()
        try:
            if self._should_die:
                self.connect_aborted('aborted')
                raise SystemExit
        finally:
            self.lock.release()
        
    @abortable
    def stop_wpa(self, iface):
        """ Stops wpa_supplicant. """
        print 'Stopping wpa_supplicant'
        iface.StopWPA()
        
    @abortable
    def put_iface_up(self, iface):
        """ Bring up given interface. """
        print 'Putting interface up...'
        self.SetStatus('interface_up')
        iface.Up()
        for x in range(0, 5):
            time.sleep(2)
            if iface.IsUp():
                return
            self.abort_if_needed()
         
        # If we get here, the interface never came up
        print "WARNING: Timed out waiting for interface to come up"


class Wireless(Controller):
    """ A wrapper for common wireless interface functions. """

    def __init__(self, debug=False):
        """ Initialize the class. """
        Controller.__init__(self, debug=debug)
        self._wpa_driver = None
        self._wireless_interface = None
        self.wiface = None 
        self.should_verify_ap = True
        
    def set_wireless_iface(self, value):
        self._wireless_interface = value
        if self.wiface:
            self.wiface.SetInterface(value)
    def get_wireless_iface(self): return self._wireless_interface
    wireless_interface = property(get_wireless_iface, set_wireless_iface) 
        
    def set_wpa_driver(self, value):
        self._wpa_driver = value
        if self.wiface:
            self.SetWPADriver(value)
    def get_wpa_driver(self): return self._wpa_driver
    wpa_driver = property(get_wpa_driver, set_wpa_driver)
    
    def set_iface(self, value):
        self.wiface = value
    def get_iface(self):
        return self.wiface
    iface = property(get_iface, set_iface)
        
    def LoadBackend(self, backend):
        """ Load a given backend. 

        Load up a backend into the backend manager and associate with
        the networking interface.
        
        """
        Controller.LoadBackend(self, backend)
        if self._backend:
            self.wiface = self._backend.WirelessInterface(self.wireless_interface,
                                                    self.debug, self.wpa_driver)

    def Scan(self, essid=None):
        """ Scan for available wireless networks.

        Keyword arguments:
        essid -- The essid of a hidden network

        Returns:
        A list of available networks sorted by strength.

        """
        def comp(x, y):
            if 'quality' in x:
                key = 'quality'
            else:
                key = 'strength'
            return cmp(x[key], y[key])
                
        if not self.wiface: return []
        wiface = self.wiface

        # Prepare the interface for scanning
        wiface.Up()

        # If there is a hidden essid then set it now, so that when it is
        # scanned it will be recognized.
        essid = misc.Noneify(essid)
        if essid is not None:
            print 'Setting hidden essid' + essid
            wiface.SetEssid(essid)
            # sleep for a bit; scanning to fast will result in nothing
            time.sleep(1)

        aps = wiface.GetNetworks()
        aps.sort(cmp=comp, reverse=True)
        
        return aps

    def Connect(self, network, debug=False):
        """ Spawn a connection thread to connect to the network.

        Keyword arguments:
        network -- network to connect to

        """
        if not self.wiface: return False
        
        self.connecting_thread = WirelessConnectThread(network,
            self.wireless_interface, self.wpa_driver, self.before_script,
            self.after_script, self.pre_disconnect_script,
            self.post_disconnect_script, self.global_dns_1,
            self.global_dns_2, self.global_dns_3, self.global_dns_dom,
            self.global_search_dom, self.wiface, self.should_verify_ap, debug)
        self.connecting_thread.setDaemon(True)
        self.connecting_thread.start()
        return True

    def GetSignalStrength(self, iwconfig=""):
        """ Get signal strength of the current network.

        Returns:
        The current signal strength.

        """
        return self.wiface.GetSignalStrength(iwconfig)

    def GetDBMStrength(self, iwconfig=""):
        """ Get the dBm signal strength of the current network.

        Returns:
        The current dBm signal strength.

        """
        return self.wiface.GetDBMStrength(iwconfig)

    def GetCurrentNetwork(self, iwconfig=""):
        """ Get current network name.

        Returns:
        The name of the currently connected network.

        """
        if self.connecting_thread and self.connecting_thread.is_connecting:
            return self.connecting_thread.network['essid']
        return self.wiface.GetCurrentNetwork(iwconfig)
    
    def GetBSSID(self):
        """ Get the BSSID of the current access point. 
        
        Returns:
        The MAC Adress of the active access point as a string, or
        None the BSSID can't be found.
        
        """
        return self.wiface.GetBSSID()

    def GetCurrentBitrate(self, iwconfig):
        """ Get the current bitrate of the interface. 
        
        Returns:
        The bitrate of the active access point as a string, or
        None the bitrate can't be found.
        
        """
        return self.wiface.GetCurrentBitrate(iwconfig)

    def GetOperationalMode(self, iwconfig):
        """ Get the current operational mode of the interface. 
        
        Returns:
        The operational mode of the interface as a string, or
        None if the operational mode can't be found.
        
        """
        return self.wiface.GetOperationalMode(iwconfig)

    def GetAvailableAuthMethods(self, iwlistauth):
        """ Get the available authentication methods for the interface. 
        
        Returns:
        The available authentication methods of the interface as a string, or
        None if the auth methods can't be found.
        
        """
        return self.wiface.GetAvailableAuthMethods(iwlistauth)

    def GetIwconfig(self):
        """ Get the out of iwconfig. """
        return self.wiface.GetIwconfig()
    
    def GetWpaSupplicantDrivers(self):
        """ Returns all valid wpa_supplicant drivers on the system. """
        return BACKEND.GetWpaSupplicantDrivers()
    
    def StopWPA(self):
        return self.wiface.StopWPA()

    def CreateAdHocNetwork(self, essid, channel, ip, enctype, key,
            enc_used):
        """ Create an ad-hoc wireless network.

        Keyword arguments:
        essid -- essid of the ad-hoc network
        channel -- channel of the ad-hoc network
        ip -- ip of the ad-hoc network
        enctype -- unused
        key -- key of the ad-hoc network
        enc_used -- encrytion enabled on ad-hoc network

        """
        wiface = self.wiface
        print 'Creating ad-hoc network'
        print 'Stopping dhcp client and wpa_supplicant'
        wiface.ReleaseDHCP()
        wiface.StopWPA()
        print 'Putting wireless interface down'
        wiface.Down()
        print 'Setting mode, channel, and essid'
        wiface.SetMode('ad-hoc')
        wiface.SetChannel(channel)
        wiface.SetEssid(essid)
        # Right now it just assumes you're using WEP
        if enc_used:
            print 'Setting encryption key'
            wiface.SetKey(key)
        print 'Putting interface up'
        wiface.Up()
        print 'Setting IP address'
        wiface.SetAddress(ip, '255.255.255.0')

    def DetectWirelessInterface(self):
        """ Detect available wireless interfaces.

        Returns:
        The first available wireless interface.

        """
        ifaces = BACKEND.GetWirelessInterfaces()
        if ifaces:
            return ifaces[0]
        else:
            return None

    def GetKillSwitchStatus(self):
        """ Get the current status of the Killswitch. 
        
        Returns:
        True if the killswitch is on, False otherwise.
        
        """
        return self.wiface.GetKillSwitchStatus()

    def SwitchRfKill(self):
        """ Switches the rfkill on/off for wireless cards. """
        types = ['wifi', 'wlan', 'wimax', 'wwan']
        try:
            if self.GetRfKillStatus():
                action = 'unblock'
            else:
                action = 'block'
            for t in types:
                cmd = ['rfkill', action, t]
                print "rfkill: %sing %s" % (action, t)
                misc.Run(cmd)
            return True
        except Exception, e:
            raise e
            return False

    def GetRfKillStatus(self):
        """ Determines if rfkill switch is active or not.

        Returns:
        True if rfkill (soft-)switch is enabled.
        """
        cmd = 'rfkill list'
        rfkill_out = misc.Run(cmd)
        soft_blocks = filter(lambda x: x.startswith('Soft'), rfkill_out.split('\t'))
        for line in map(lambda x: x.strip(), soft_blocks):
            if line.endswith('yes'):
                return True
        return False

    def Disconnect(self):
        """ Disconnect the given iface.
        
        Executes the disconnect script associated with a given interface,
        Resets it's IP address, and puts the interface down then up.
        
        """
        if BACKEND.NeedsExternalCalls():
            iwconfig = self.GetIwconfig()
        else:
            iwconfig = None
        bssid = self.wiface.GetBSSID(iwconfig)
        essid = self.wiface.GetCurrentNetwork(iwconfig) 

        Controller.Disconnect(self, 'wireless', essid, bssid)
        self.StopWPA()
    
    def SetWPADriver(self, driver):
        """ Sets the wpa_supplicant driver associated with the interface. """
        self.wiface.SetWpaDriver(driver)

 
class WirelessConnectThread(ConnectThread):
    """ A thread class to perform the connection to a wireless network.

    This thread, when run, will perform the necessary steps to connect
    to the specified network.

    """

    def __init__(self, network, wireless, wpa_driver, before_script,
                 after_script, pre_disconnect_script, post_disconnect_script,
                 gdns1, gdns2, gdns3, gdns_dom, gsearch_dom, wiface, 
                 should_verify_ap, debug=False):
        """ Initialise the thread with network information.

        Keyword arguments:
        network -- the network to connect to
        wireless -- name of the wireless interface
        wpa_driver -- type of wireless interface
        before_script -- script to run before bringing up the interface
        after_script -- script to run after bringing up the interface
        pre_disconnect_script -- script to run before disconnection
        post_disconnect_script -- script to run after disconnection
        gdns1 -- global DNS server 1
        gdns2 -- global DNS server 2
        gdns3 -- global DNS server 3

        """
        ConnectThread.__init__(self, network, wireless, before_script, 
                               after_script, pre_disconnect_script,
                               post_disconnect_script, gdns1, gdns2,
                               gdns3, gdns_dom, gsearch_dom, wiface, debug)
        self.wpa_driver = wpa_driver
        self.should_verify_ap = should_verify_ap


    def _connect(self):
        """ The main function of the connection thread.

        This function performs the necessary calls to connect to the
        specified network, using the information provided. The following
        indicates the steps taken.
        1. Run pre-connection script.
        2. Take down the interface and clean up any previous
           connections.
        3. Generate a PSK if required and authenticate.
        4. Associate with the WAP.
        5. Get/set IP address and DNS servers.

        """
        wiface = self.iface
        self.is_connecting = True
        
        # Run pre-connection script.
        self.run_global_scripts_if_needed(wpath.preconnectscripts,
                                          extra_parameters=('wireless',
                                                    self.network['essid'],
                                                    self.network['bssid'])
                                         )
        self.run_script_if_needed(self.before_script, 'pre-connection', 
                                  self.network['bssid'], self.network['essid'])

        
        # Take down interface and clean up previous connections.
        self.put_iface_down(wiface)
        self.release_dhcp_clients(wiface)
        self.reset_ip_addresses(wiface)
        self.stop_wpa(wiface)
        self.flush_routes(wiface)
        wiface.SetMode(self.network['mode'])

        # Put interface up.
        self.SetStatus('configuring_interface')
        self.put_iface_up(wiface)
        
        # Generate PSK and authenticate if needed.
        if self.wpa_driver != 'ralink legacy':
            self.generate_psk_and_authenticate(wiface)
            
        # Associate.
        wiface.Associate(self.network['essid'], self.network['channel'],
                         self.network['bssid'])

        # Authenticate after association for Ralink legacy cards.
        if self.wpa_driver == 'ralink legacy':
            if self.network.get('key'):
                wiface.Authenticate(self.network)
                
        # Validate Authentication.
        if self.network.get('enctype'):
            self.SetStatus('validating_authentication')
            if not wiface.ValidateAuthentication(time.time()):
                print "connect result is %s" % self.connect_result
                if not self.connect_result or self.connect_result == 'failed':
                    self.abort_connection('bad_pass')

        # Set up gateway, IP address, and DNS servers.
        self.set_broadcast_address(wiface)
        self.set_ip_address(wiface)
        self.set_dns_addresses(wiface)
        self.verify_association(wiface)
        
        # Run post-connection script.
        self.run_global_scripts_if_needed(wpath.postconnectscripts,
                                          extra_parameters=('wireless',
                                                    self.network['essid'],
                                                    self.network['bssid'])
                                         )
        self.run_script_if_needed(self.after_script, 'post-connection', 
                                  self.network['bssid'], self.network['essid'])

        self.SetStatus('done')
        print 'Connecting thread exiting.'
        if self.debug:
            print "IP Address is: " + str(wiface.GetIP())
        self.connect_result = "success"
        self.is_connecting = False
        
    @abortable
    def verify_association(self, iface):
        """ Verify that our association the AP is valid.
        
        Try to ping the gateway we have set to see if we're
        really associated with it.  This is only done if
        we're using a static IP.
        
        """
        if self.network.get('gateway') and self.should_verify_ap:
            self.SetStatus('verifying_association')
            print "Verifying AP association..."
            for tries in range(1, 11):
                print "Attempt %d of 10..." % tries
                retcode = self.iface.VerifyAPAssociation(self.network['gateway'])
                if retcode == 0: 
                    print "Successfully associated."
                    break
                time.sleep(1)
            #TODO this should be in wnettools.py
            if retcode:
                print "Connection Failed: Failed to ping the access point!"
                # Clean up before aborting.
                iface.SetAddress('0.0.0.0')
                iface.FlushRoutes()
                if hasattr(iface, "StopWPA"):
                    iface.StopWPA()
                self.abort_connection('association_failed')
        else:
            print 'not verifying'
        
    @abortable
    def generate_psk_and_authenticate(self, wiface):
        """ Generates a PSK and authenticates if necessary. 
        
        Generates a PSK, and starts the authentication process
        if encryption is on.
        
        """
        # Check to see if we need to generate a PSK (only for non-ralink
        # cards).
        if self.debug:
            print "enctype is %s" % self.network.get('enctype')
        if self.network.get('key') and 'wpa' in str(self.network.get('enctype')):
            self.SetStatus('generating_psk')
            print 'Generating psk...'
            self.network['psk'] = wiface.GeneratePSK(self.network)
            
            if not self.network.get('psk'):
                self.network['psk'] = self.network['key']
                print 'WARNING: PSK generation failed!  Falling back to ' + \
                      'wireless key.\nPlease report this error to the wicd ' + \
                      'developers!'
        # Generate the wpa_supplicant file...
        if self.network.get('enctype'):
            self.SetStatus('generating_wpa_config')
            print 'Attempting to authenticate...'
            wiface.Authenticate(self.network)


class Wired(Controller):
    """ A wrapper for common wired interface functions. """

    def __init__(self, debug=False):
        """ Initialise the class. """
        Controller.__init__(self, debug=debug)
        self.wpa_driver = None
        self._link_detect = None
        self._wired_interface = None
        self.liface = None
        
    def set_link_detect(self, value):
        self._link_detect = value
        if self.liface:
            self.liface.link_detect = value
    def get_link_detect(self): return self._link_detect
    link_detect = property(get_link_detect, set_link_detect)
    
    
    def set_wired_iface(self, value):
        self._wired_interface = value
        if self.liface:
            self.liface.SetInterface(value)            
    def get_wired_iface(self): return self._wired_interface
    wired_interface = property(get_wired_iface, set_wired_iface)
    
    def set_iface(self, value):
        self.liface = value
    def get_iface(self):
        return self.liface
    iface = property(get_iface, set_iface)
    
    def LoadBackend(self, backend):
        """ Load the backend up. """
        Controller.LoadBackend(self, backend)
        if self._backend:
            self.liface = self._backend.WiredInterface(self.wired_interface,
                                                       self.debug)

    def CheckPluggedIn(self):
        """ Check whether the wired connection is plugged in.

        Returns:
        The status of the physical connection link.

        """
        return self.liface.GetPluggedIn()

    def Connect(self, network, debug=False):
        """ Spawn a connection thread to connect to the network.

        Keyword arguments:
        network -- network to connect to

        """
        if not self.liface: return False
        self.connecting_thread = WiredConnectThread(network,
            self.wired_interface, self.before_script, self.after_script,
            self.pre_disconnect_script, self.post_disconnect_script,
            self.global_dns_1, self.global_dns_2, self.global_dns_3,
            self.global_dns_dom, self.global_search_dom, self.liface,
            debug)
        self.connecting_thread.setDaemon(True)
        self.connecting_thread.start()
        return self.connecting_thread
    
    def Disconnect(self):
        Controller.Disconnect(self, 'wired', 'wired', 'wired')
        self.StopWPA()
    
    def StopWPA(self):
        self.liface.StopWPA()
    
    def DetectWiredInterface(self):
        """ Attempts to automatically detect a wired interface. """
        try:
            return BACKEND.GetWiredInterfaces()[0]
        except IndexError:
            return None


class WiredConnectThread(ConnectThread):
    """ A thread class to perform the connection to a wired network.

    This thread, when run, will perform the necessary steps to connect
    to the specified network.

    """
    def __init__(self, network, wired, before_script, after_script, 
                 pre_disconnect_script, post_disconnect_script, gdns1,
                 gdns2, gdns3, gdns_dom, gsearch_dom, liface, debug=False):
        """ Initialise the thread with network information.

        Keyword arguments:
        network -- the network to connect to
        wireless -- name of the wireless interface
        wired -- name of the wired interface
        before_script -- script to run before bringing up the interface
        after_script -- script to run after bringing up the interface
        pre_disconnect_script -- script to run before disconnection
        post_disconnect_script -- script to run after disconnection
        gdns1 -- global DNS server 1
        gdns2 -- global DNS server 2
        gdns3 -- global DNS server 3

        """
        ConnectThread.__init__(self, network, wired, before_script, 
                               after_script, pre_disconnect_script,
                               post_disconnect_script, gdns1, gdns2,
                               gdns3, gdns_dom, gsearch_dom, liface,
                               debug)

    def _connect(self):
        """ The main function of the connection thread.

        This function performs the necessary calls to connect to the
        specified network, using the information provided. The following
        indicates the steps taken.
        1. Run pre-connection script.
        2. Take down the interface and clean up any previous
           connections.
        3. Bring up the interface.
        4. Get/set IP address and DNS servers.
        5. Run post-connection script.

        """
        liface = self.iface

        self.is_connecting = True

        # Run pre-connection script.
        self.run_global_scripts_if_needed(wpath.preconnectscripts,
                                          extra_parameters=('wired', 'wired',
                                                            self.network['profilename'])
                                          )
        self.run_script_if_needed(self.before_script, 'pre-connection', 'wired', 
                                  'wired')

        # Take down interface and clean up previous connections.
        self.put_iface_down(liface)
        self.release_dhcp_clients(liface)
        self.reset_ip_addresses(liface)
        self.stop_wpa(liface)
        self.flush_routes(liface)
        
        # Bring up interface.
        self.put_iface_up(liface)
        
        # Manage encryption.
        if self.network.get('encryption_enabled'):
            liface.Authenticate(self.network)
        
        # Set gateway, IP adresses, and DNS servers.
        self.set_broadcast_address(liface)
        self.set_ip_address(liface)
        self.set_dns_addresses(liface)
        
        # Run post-connection script.
        self.run_global_scripts_if_needed(wpath.postconnectscripts,
                                          extra_parameters=('wired', 'wired',
                                                self.network['profilename'])
                                         )
        self.run_script_if_needed(self.after_script, 'post-connection', 'wired', 
                                  'wired')

        self.SetStatus('done')
        print 'Connecting thread exiting.'
        if self.debug:
            print "IP Address is: " + str(liface.GetIP())
        
        self.connect_result = "success"
        self.is_connecting = False



""" guiutil - A collection of commonly used gtk/gui functions and classes. """
#
#   Copyright (C) 2007 - 2009 Adam Blackburn
#   Copyright (C) 2007 - 2009 Dan O'Reilly
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License Version 2 as
#   published by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import gtk
import os.path

import wicd.wpath as wpath

HAS_NOTIFY = True
try:
    import pynotify
    if not pynotify.init("Wicd"):
        print 'Could not initalize pynotify'
        HAS_NOTIFY = False
except ImportError:
    print "Importing pynotify failed, notifications disabled."
    HAS_NOTIFY = False
 
print "Has notifications support", HAS_NOTIFY

if wpath.no_use_notifications:
    print 'Notifications disabled during setup.py configure'
    
def can_use_notify():
    use_notify = os.path.exists(os.path.join(os.path.expanduser('~/.wicd'),
                                             'USE_NOTIFICATIONS')
                                )
    return use_notify and HAS_NOTIFY and not wpath.no_use_notifications
    
def error(parent, message, block=True): 
    """ Shows an error dialog. """
    def delete_event(dialog, id):
        dialog.destroy()
    if can_use_notify() and not block:
        notification = pynotify.Notification("ERROR", message, "error")
        notification.show()
        return
    dialog = gtk.MessageDialog(parent, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR,
                               gtk.BUTTONS_OK)
    dialog.set_markup(message)
    if not block:
        dialog.present()
        dialog.connect("response", delete_event)
    else:
        dialog.run()
        dialog.destroy()
    
def alert(parent, message, block=True): 
    """ Shows an warning dialog. """
    def delete_event(dialog, id):
        dialog.destroy()
    dialog = gtk.MessageDialog(parent, gtk.DIALOG_MODAL, gtk.MESSAGE_WARNING,
                               gtk.BUTTONS_OK)
    dialog.set_markup(message)
    if not block:
        dialog.present()
        dialog.connect("response", delete_event)
    else:
        dialog.run()
        dialog.destroy()

def string_input(prompt, secondary, textbox_label):
    # based on a version of a PyGTK text entry from
    # http://ardoris.wordpress.com/2008/07/05/pygtk-text-entry-dialog/

    def dialog_response(entry, dialog, response):
        dialog.response(response)

    dialog = gtk.MessageDialog(
        None,
        gtk.DIALOG_MODAL,
        gtk.MESSAGE_QUESTION,
        gtk.BUTTONS_OK_CANCEL,
        None)

    # set the text
    dialog.set_markup("<span size='larger'><b>" + prompt + "</b></span>")
    # add the secondary text
    dialog.format_secondary_markup(secondary)

    entry = gtk.Entry()
    # allow the user to press enter instead of clicking OK
    entry.connect("activate", dialog_response, dialog, gtk.RESPONSE_OK)

    # create an hbox and pack the label and entry in
    hbox = gtk.HBox()
    hbox.pack_start(gtk.Label(textbox_label), False, 4, 4)
    hbox.pack_start(entry)

    # pack the boxes and show the dialog
    dialog.vbox.pack_end(hbox, True, True, 0)
    dialog.show_all()

    if dialog.run() == gtk.RESPONSE_OK:
        text = entry.get_text()
        dialog.destroy()
        return text
    else:
        dialog.destroy()
        return None

class SmallLabel(gtk.Label):
    def __init__(self, text=''):
        gtk.Label.__init__(self, text)
        self.set_size_request(50, -1)
        
class LeftAlignedLabel(gtk.Label):
    def __init__(self, label=None):
        gtk.Label.__init__(self, label)
        self.set_alignment(0.0, 0.5)

class LabelEntry(gtk.HBox):
    """ A label on the left with a textbox on the right. """
    def __init__(self,text):
        gtk.HBox.__init__(self)
        self.entry = gtk.Entry()
        self.entry.set_size_request(200, -1)
        self.label = LeftAlignedLabel()
        self.label.set_text(text)
        self.label.set_size_request(170, -1)
        self.pack_start(self.label, fill=True, expand=True)
        self.pack_start(self.entry, fill=False, expand=False)
        self.label.show()
        self.entry.show()
        self.entry.connect('focus-out-event', self.hide_characters)
        self.entry.connect('focus-in-event', self.show_characters)
        self.auto_hide_text = False
        self.show()

    def set_text(self, text):
        # For compatibility...
        self.entry.set_text(text)

    def get_text(self):
        return self.entry.get_text()

    def set_auto_hidden(self, value):
        self.entry.set_visibility(False)
        self.auto_hide_text = value

    def show_characters(self, widget=None, event=None):
        # When the box has focus, show the characters
        if self.auto_hide_text and widget:
            self.entry.set_visibility(True)

    def set_sensitive(self, value):
        self.entry.set_sensitive(value)
        self.label.set_sensitive(value)

    def hide_characters(self, widget=None, event=None):
        # When the box looses focus, hide them
        if self.auto_hide_text and widget:
            self.entry.set_visibility(False)


class GreyLabel(gtk.Label):
    """ Creates a grey gtk.Label. """
    def __init__(self):
        gtk.Label.__init__(self)

    def set_label(self, text):
        self.set_markup(text)
        self.set_alignment(0, 0)


class ProtectedLabelEntry(gtk.HBox):
    """ A LabelEntry with a CheckButton that protects the entry text. """
    def __init__(self, label_text):
        gtk.HBox.__init__(self)
        self.entry = gtk.Entry()
        self.entry.set_size_request(200, -1)
        self.entry.set_visibility(False)
        self.label = LeftAlignedLabel()
        self.label.set_text(label_text)
        self.label.set_size_request(165, -1)
        self.check = gtk.CheckButton()
        self.check.set_size_request(5, -1)
        self.check.set_active(False)
        self.check.set_focus_on_click(False)
        self.pack_start(self.label, fill=True, expand=True)
        self.pack_start(self.check, fill=True, expand=True)
        self.pack_start(self.entry, fill=False, expand=False)
        self.label.show()
        self.check.show()
        self.entry.show()
        self.check.connect('clicked', self.click_handler)
        self.show()

    def set_entry_text(self, text):
        # For compatibility...
        self.entry.set_text(text)

    def get_entry_text(self):
        return self.entry.get_text()

    def set_sensitive(self, value):
        self.entry.set_sensitive(value)
        self.label.set_sensitive(value)
        self.check.set_sensitive(value)

    def click_handler(self, widget=None, event=None):
        active = self.check.get_active()
        self.entry.set_visibility(active)



#!/usr/bin/env python

"""configscript_curses.py
Kind of like configscript.py, except written using urwid.

Also recycles a lot of configscript.py, too. :-)
"""

#       Copyright (C) 2009 Andrew Psaltis

#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

from wicd import misc
from wicd.translations import _
import configscript
from configscript import write_scripts,get_script_info,get_val,none_to_blank,blank_to_none

import urwid
import urwid.curses_display
import sys
import os

def main(argv):
    global ui,frame
    if len(argv) < 2:
        print 'Network id to configure is missing, aborting.'
        sys.exit(1)

    ui = urwid.curses_display.Screen()
    ui.register_palette( [
        ('body','default','default'),
        ('focus','dark magenta','light gray'),
        ('editcp', 'default', 'default', 'standout'),
        ('editbx', 'light gray', 'dark blue'),
        ('editfc', 'white','dark blue', 'bold')] )

    network = argv[1]
    network_type = argv[2]
    
    script_info = get_script_info(network, network_type)

    blank = urwid.Text('')
    pre_entry_t = ('body',_('Pre-connection Script')+': ')
    post_entry_t  = ('body',_('Post-connection Script')+': ')
    pre_disconnect_entry_t = ('body',_('Pre-disconnection Script')+': ')
    post_disconnect_entry_t = ('body',_('Post-disconnection Script')+': ')

    global pre_entry,post_entry,pre_disconnect_entry,post_disconnect_entry
    pre_entry  = urwid.AttrWrap(urwid.Edit(pre_entry_t,
            none_to_blank(script_info.get('pre_entry'))),'editbx','editfc' )
    post_entry  = urwid.AttrWrap(urwid.Edit(post_entry_t,
            none_to_blank(script_info.get('post_entry'))),'editbx','editfc' )

    pre_disconnect_entry  = urwid.AttrWrap(urwid.Edit(pre_disconnect_entry_t,
            none_to_blank(script_info.get('pre_disconnect_entry'))),'editbx','editfc' )
    post_disconnect_entry  = urwid.AttrWrap(urwid.Edit(post_disconnect_entry_t,
            none_to_blank(script_info.get('post_disconnect_entry'))),'editbx','editfc' )

    # The buttons
    ok_button = urwid.AttrWrap(urwid.Button(_('OK'),ok_callback),'body','focus')
    cancel_button = urwid.AttrWrap(urwid.Button(_('Cancel'),cancel_callback),'body','focus')

    button_cols = urwid.Columns([ok_button,cancel_button],dividechars=1)

    lbox = urwid.Pile([('fixed',2,urwid.Filler(pre_entry)),
                       #('fixed',urwid.Filler(blank),1),
                       ('fixed',2,urwid.Filler(post_entry)),
                       ('fixed',2,urwid.Filler(pre_disconnect_entry)),
                       ('fixed',2,urwid.Filler(post_disconnect_entry)),
                       #blank,blank,blank,blank,blank,
                       urwid.Filler(button_cols,'bottom')
                       ])
    frame = urwid.Frame(lbox)
    result = ui.run_wrapper(run)
 
    if result == True:
        script_info["pre_entry"] = blank_to_none(pre_entry.get_edit_text())
        script_info["post_entry"] = blank_to_none(post_entry.get_edit_text())
        script_info["pre_disconnect_entry"] = blank_to_none(pre_disconnect_entry.get_edit_text())
        script_info["post_disconnect_entry"] = blank_to_none(post_disconnect_entry.get_edit_text())
        write_scripts(network, network_type, script_info)

OK_PRESSED = False
CANCEL_PRESSED = False
def ok_callback(button_object,user_data=None):
    global OK_PRESSED
    OK_PRESSED = True
def cancel_callback(button_object,user_data=None):
    global CANCEL_PRESSED
    CANCEL_PRESSED = True
def run():
    dim = ui.get_cols_rows()
    ui.set_mouse_tracking()

    keys = True
    while True:
        if keys:
            ui.draw_screen(dim, frame.render(dim, True))
        keys = ui.get_input()

        if "window resize" in keys:
            dim = ui.get_cols_rows()
        if "esc" in keys or 'Q' in keys:
            return False
        for k in keys:
            #Send key to underlying widget:
            if urwid.is_mouse_event(k):
                event, button, col, row = k
                frame.mouse_event( dim,
                        event, button, col, row,
                        focus=True)
            else:
                frame.keypress(dim, k)
            # Check if buttons are pressed.
        if CANCEL_PRESSED:
            return False
        if OK_PRESSED or 'meta enter' in keys:
            return True

if __name__ == '__main__':
    if os.getuid() != 0:
        print "Root privileges are required to configure scripts.  Exiting."
        sys.exit(0)
    main(sys.argv)



#!/usr/bin/env python

"""prefs_curses.py -- Pretty, tabbable, console preferences dialog"""

#       Copyright (C) 2008-2009 Andrew Psaltis

#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 2 of the License, or
#       (at your option) any later version.
#       
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#       
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.

import urwid
import urwid.curses_display

from wicd import misc
from wicd import dbusmanager
from wicd.translations import _
from curses_misc import SelText,DynWrap,DynRadioButton,ComboBox,TabColumns

daemon = None
wireless = None
wired = None

from wicd.translations import language

class PrefsDialog(urwid.WidgetWrap):
    def __init__(self, body, pos, ui, dbus=None):
        global daemon, wireless, wired

        daemon = dbus['daemon']
        wireless = dbus['wireless']
        wired = dbus['wired']

        width,height =  ui.get_cols_rows()
        height -= 3
        #width = 80
        #height = 20
        # Stuff that goes at the top

        header0_t = _('General Settings')
        header1_t = _('External Programs')
        header2_t = _('Advanced Settings')
        self.header0 = urwid.AttrWrap(SelText(header0_t), 'tab active', 'focus')
        self.header1 = urwid.AttrWrap(SelText(header1_t), 'body', 'focus')
        self.header2 = urwid.AttrWrap(SelText(header2_t), 'body', 'focus')
        title   = _('Preferences')

        # Blank line
        _blank = urwid.Text('')

        ####
        #### Text in the widgets
        ####

        # General Settings
        net_cat_t           = ('header', _('Network Interfaces'))
        wired_t             = ('editcp', _('Wired Interface')+':   ')
        wless_t             = ('editcp', _('Wireless Interface')+':')
        always_show_wired_t = _('''Always show wired interface''')
        prefer_wired_t      = _('''Always switch to wired connection when available''')

        global_dns_cat_t = ('header', _('Global DNS servers'))
        global_dns_t     = ('editcp', _('Use global DNS servers'))
        dns_dom_t        = ('editcp', '    '+_('DNS domain')+':   ')
        search_dom_t     = ('editcp', '    '+_('Search domain')+':')
        dns1_t           = ('editcp', '    '+_('DNS server')+' 1: ')
        dns2_t           = ('editcp', '    '+_('DNS server')+' 2: ')
        dns3_t           = ('editcp', '    '+_('DNS server')+' 3: ')


        wired_auto_cat_t= ('header', _('Wired Autoconnect Settings'))
        wired_auto_1_t = _('Use default profile on wired autoconnect')
        wired_auto_2_t = _('Prompt for profile on wired autoconnect')
        wired_auto_3_t = _('Use last used profile on wired autoconnect')

        auto_reconn_cat_t = ('header', _('Automatic Reconnection'))
        auto_reconn_t = _('Automatically reconnect on connection loss')

        #### External Programs
        automatic_t = _('Automatic (recommended)')

        dhcp_header_t = ('header', _('DHCP Client'))
        # Automatic
        dhcp1_t  = 'dhclient'
        dhcp2_t  = 'dhcpcd'
        dhcp3_t  = 'pump'
        dhcp4_t  = 'udhcpc'

        wired_detect_header_t = ('header', _('Wired Link Detection'))
        wired1_t              = 'ethtool'
        wired2_t              = 'mii-tool'

        flush_header_t = ('header', _('Route Table Flushing'))
        flush1_t           = 'ip'
        flush2_t           = 'route'
 
        #### Advanced Settings
        wpa_cat_t=('header', _('WPA Supplicant'))
        wpa_t=('editcp','Driver:')
        wpa_list = []
        wpa_warn_t = ('important', _('You should almost always use wext as the WPA supplicant driver'))
        
        backend_cat_t = ('header', _('Backend'))
        backend_t = _('Backend')+':'
        backend_list = []
        
        debug_cat_t = ('header', _('Debugging'))
        debug_mode_t = _('Enable debug mode')

        wless_cat_t = ('header', _('Wireless Interface'))
        use_dbm_t = _('Use dBm to measure signal strength')
        verify_ap_t = _('Ping static gateways after connecting to verify association')
        


        ####
        #### UI Widgets
        ####

        # General Settings
        self.net_cat    = urwid.Text(net_cat_t)
        self.wired_edit = urwid.AttrWrap(urwid.Edit(wired_t),'editbx','editfc')
        self.wless_edit = urwid.AttrWrap(urwid.Edit(wless_t),'editbx','editfc')
        self.prefer_wired_chkbx = urwid.CheckBox(prefer_wired_t)
        self.global_dns_cat = urwid.Text(global_dns_cat_t)
        # Default the global DNS settings to off.  They will be reenabled later
        # if so required.
        global_dns_state = False
        self.global_dns_checkb  = urwid.CheckBox(global_dns_t, global_dns_state,
                on_state_change=self.global_dns_trigger)
        self.search_dom = DynWrap(urwid.Edit(search_dom_t), global_dns_state)
        self.dns_dom    = DynWrap(urwid.Edit(dns_dom_t), global_dns_state)
        self.dns1       = DynWrap(urwid.Edit(dns1_t), global_dns_state)
        self.dns2       = DynWrap(urwid.Edit(dns2_t), global_dns_state)
        self.dns3       = DynWrap(urwid.Edit(dns3_t), global_dns_state)


        self.always_show_wired_checkb = urwid.CheckBox(always_show_wired_t)

        self.wired_auto_l  = []
        self.wired_auto_cat= urwid.Text(wired_auto_cat_t)
        self.wired_auto_1  = urwid.RadioButton(self.wired_auto_l,wired_auto_1_t)
        self.wired_auto_2  = urwid.RadioButton(self.wired_auto_l,wired_auto_2_t)
        self.wired_auto_3  = urwid.RadioButton(self.wired_auto_l,wired_auto_3_t)

        self.auto_reconn_cat    = urwid.Text(auto_reconn_cat_t)
        self.auto_reconn_checkb = urwid.CheckBox(auto_reconn_t)
        generalLB = urwid.ListBox([self.net_cat,
                                   self.wless_edit,#_blank,
                                   self.wired_edit,
                                   self.always_show_wired_checkb,
                                   self.prefer_wired_chkbx,_blank,
                                   self.global_dns_cat,
                                   self.global_dns_checkb,#_blank,
                                   self.search_dom,self.dns_dom,
                                   self.dns1,self.dns2,self.dns3,_blank,
                                   self.wired_auto_cat,
                                   self.wired_auto_1,
                                   self.wired_auto_2,
                                   self.wired_auto_3, _blank,
                                   self.auto_reconn_cat,
                                   self.auto_reconn_checkb
                                  ])

        #### External Programs tab
        automatic_t = _('Automatic (recommended)')

        self.dhcp_header = urwid.Text(dhcp_header_t)
        self.dhcp_l = []

        # Order of these is flipped in the actual interface,
        # (2,3,1 -> dhcpcd, pump, dhclient), because dhclient often doesn't like
        # to work on several distros.
        self.dhcp0  = urwid.RadioButton(self.dhcp_l ,automatic_t)
        self.dhcp1  = DynRadioButton(self.dhcp_l, dhcp1_t)
        self.dhcp2  = DynRadioButton(self.dhcp_l, dhcp2_t)
        self.dhcp3  = DynRadioButton(self.dhcp_l, dhcp3_t)
        self.dhcp4  = DynRadioButton(self.dhcp_l, dhcp4_t)
        self.dhcp_l = [self.dhcp0,self.dhcp1,self.dhcp2,self.dhcp3,self.dhcp4]

        self.wired_l = []
        self.wired_detect_header = urwid.Text(wired_detect_header_t)
        self.wired0         = urwid.RadioButton(self.wired_l, automatic_t)
        self.wired1         = DynRadioButton(self.wired_l, wired1_t)
        self.wired2         = DynRadioButton(self.wired_l, wired2_t)
        self.wired_l = [self.wired0, self.wired1, self.wired2]

        self.flush_l = []
        self.flush_header   = urwid.Text(flush_header_t)
        self.flush0         = urwid.RadioButton(self.flush_l,automatic_t)
        self.flush1         = DynRadioButton(self.flush_l,flush1_t)
        self.flush2         = DynRadioButton(self.flush_l,flush2_t)
        self.flush_l = [self.flush0,self.flush1,self.flush2]

        externalLB = urwid.ListBox([self.dhcp_header,
                                    self.dhcp0,self.dhcp2,self.dhcp3,self.dhcp1,
                                    self.dhcp4,
                                    _blank,
                                    self.wired_detect_header,
                                    self.wired0,self.wired1,self.wired2,
                                    _blank,
                                    self.flush_header,
                                    self.flush0,self.flush1,self.flush2
                                   ])


        #### Advanced settings
        self.wpa_cat      = urwid.Text(wpa_cat_t)
        self.wpa_cbox     = ComboBox(wpa_t)
        self.wpa_warn     = urwid.Text(wpa_warn_t)
        
        self.backend_cat  = urwid.Text(backend_cat_t)
        self.backend_cbox = ComboBox(backend_t)
        
        self.debug_cat           = urwid.Text(debug_cat_t)
        self.debug_mode_checkb   = urwid.CheckBox(debug_mode_t)

        self.wless_cat        = urwid.Text(wless_cat_t)
        self.use_dbm_checkb   = urwid.CheckBox(use_dbm_t)
        self.verify_ap_checkb = urwid.CheckBox(verify_ap_t)


        advancedLB = urwid.ListBox([self.wpa_cat,
                                    self.wpa_cbox,self.wpa_warn,_blank,
                                    self.backend_cat,
                                    self.backend_cbox,_blank,
                                    self.debug_cat,
                                    self.debug_mode_checkb, _blank,
                                    self.wless_cat,
                                    self.use_dbm_checkb, _blank,
                                    self.verify_ap_checkb, _blank
                                    ])


        headerList = [self.header0,self.header1,self.header2]
        lbList = [generalLB,externalLB,advancedLB]
        self.tab_map = {self.header0 : generalLB,
                        self.header1 : externalLB,
                        self.header2 : advancedLB}
        #self.load_settings()

        self.tabs = TabColumns(headerList,lbList,_('Preferences'))
        self.__super.__init__(self.tabs)
        
    def load_settings(self):

        ### General Settings
        # ComboBox does not like dbus.Strings as text markups.  My fault. :/
        wless_iface = unicode(daemon.GetWirelessInterface())
        wired_iface = unicode(daemon.GetWiredInterface())
        self.wless_edit.set_edit_text(wless_iface)
        self.wired_edit.set_edit_text(wired_iface)

        self.always_show_wired_checkb.set_state(
                daemon.GetAlwaysShowWiredInterface())
        self.prefer_wired_chkbx.set_state(daemon.GetPreferWiredNetwork())
        # DNS
        self.global_dns_checkb.set_state(daemon.GetUseGlobalDNS())
        theDNS = daemon.GetGlobalDNSAddresses()

        i = 0
        for w in self.dns1,self.dns2,self.dns3,self.dns_dom,self.search_dom :
            w.set_edit_text(misc.noneToBlankString(theDNS[i]))
            i+=1

        # Wired Automatic Connection
        self.wired_auto_l[daemon.GetWiredAutoConnectMethod()-1]
        self.auto_reconn_checkb.set_state(daemon.GetAutoReconnect())

        def find_avail(apps):
            for app in apps[1:]:
                app.set_sensitive(daemon.GetAppAvailable(app.get_label()))

        ### External Programs
        find_avail(self.dhcp_l)
        dhcp_method = daemon.GetDHCPClient()
        self.dhcp_l[dhcp_method].set_state(True)
        
        find_avail(self.wired_l)
        wired_link_method = daemon.GetLinkDetectionTool()
        self.wired_l[wired_link_method].set_state(True)

        find_avail(self.flush_l)
        flush_method = daemon.GetFlushTool()
        self.flush_l[flush_method].set_state(True)

        ### Advanced settings
        # wpa_supplicant janx
        self.wpadrivers = wireless.GetWpaSupplicantDrivers()
        self.wpadrivers.append("ralink_legacy")
        self.wpadrivers.append('none')
        # Same as above with the dbus.String
        self.thedrivers = [unicode(w) for w in self.wpadrivers]
        self.wpa_cbox.set_list(self.thedrivers)
        
        # Pick where to begin first:
        def_driver = daemon.GetWPADriver()
        try:
            self.wpa_cbox.set_focus(self.wpadrivers.index(def_driver))
        except ValueError:
            pass # It defaults to 0 anyway (I hope)

        self.backends = daemon.GetBackendList()
        self.thebackends= [unicode(w) for w in self.backends]
        self.backend_cbox.set_list(self.thebackends) 
        cur_backend = daemon.GetSavedBackend()
        try:
            self.backend_cbox.set_focus(self.thebackends.index(cur_backend))
        except ValueError:
            self.backend_cbox.set_focus(0)

        # Three last checkboxes
        self.debug_mode_checkb.set_state(daemon.GetDebugMode())
        self.use_dbm_checkb.set_state(daemon.GetSignalDisplayType())
        self.verify_ap_checkb.set_state(daemon.GetShouldVerifyAp())

    def save_settings(self):
        """ Pushes the selected settings to the daemon.
            This exact order is found in prefs.py"""
        daemon.SetUseGlobalDNS(self.global_dns_checkb.get_state())

        for i in [self.dns1, self.dns2,
                  self.dns3,self.dns_dom, self.search_dom, self.dns_dom]:
            i.set_edit_text(i.get_edit_text().strip())

        daemon.SetGlobalDNS(self.dns1.get_edit_text(), self.dns2.get_edit_text(),
                            self.dns3.get_edit_text(), self.dns_dom.get_edit_text(),
                            self.search_dom.get_edit_text())
        daemon.SetWirelessInterface(self.wless_edit.get_edit_text())
        daemon.SetWiredInterface(self.wired_edit.get_edit_text())
        daemon.SetWPADriver(self.wpadrivers[self.wpa_cbox.get_focus()[1]])
        daemon.SetAlwaysShowWiredInterface(self.always_show_wired_checkb.get_state())
        daemon.SetAutoReconnect(self.auto_reconn_checkb.get_state())
        daemon.SetDebugMode(self.debug_mode_checkb.get_state())
        daemon.SetSignalDisplayType(int(self.use_dbm_checkb.get_state()))
        daemon.SetShouldVerifyAp(int(self.verify_ap_checkb.get_state()))
        daemon.SetPreferWiredNetwork(bool(self.prefer_wired_chkbx.get_state()))
        if self.wired_auto_2.get_state():
            daemon.SetWiredAutoConnectMethod(2)
        elif self.wired_auto_3.get_state():
            daemon.SetWiredAutoConnectMethod(3)
        else:
            daemon.SetWiredAutoConnectMethod(1)

        daemon.SetBackend(self.backends[self.backend_cbox.get_focus()[1]])
            
        # External Programs Tab
        if self.dhcp0.get_state():
            dhcp_client = misc.AUTO
        elif self.dhcp1.get_state():
            dhcp_client = misc.DHCLIENT
        elif self.dhcp2.get_state():
            dhcp_client = misc.DHCPCD
        elif self.dhcp3.get_state():
            dhcp_client = misc.PUMP
        else:
            dhcp_client = misc.UDHCPC
        daemon.SetDHCPClient(dhcp_client)
        
        if self.wired0.get_state():
            link_tool = misc.AUTO
        elif self.wired1.get_state():
            link_tool = misc.ETHTOOL
        else:
            link_tool = misc.MIITOOL
        daemon.SetLinkDetectionTool(link_tool)
        
        if self.flush0.get_state():
            flush_tool = misc.AUTO
        elif self.flush1.get_state():
            flush_tool = misc.IP
        else:
            flush_tool = misc.ROUTE
        daemon.SetFlushTool(flush_tool)

    # DNS CheckBox callback
    def global_dns_trigger(self,check_box,new_state,user_data=None):
        for w in self.dns1,self.dns2,self.dns3,self.dns_dom,self.search_dom:
            w.set_sensitive(new_state)

    def ready_widgets(self,ui,body):
        self.wpa_cbox.build_combobox(body,ui,4)
        self.backend_cbox.build_combobox(body,ui,8)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" Network interface control tools for wicd.

This module implements functions to control and obtain information from
network interfaces.

class BaseInterface() -- Control a network interface.
class BaseWiredInterface() -- Control a wired network interface.
class BaseWirelessInterface() -- Control a wireless network interface.

"""

#
#   Copyright (C) 2007 - 2009 Adam Blackburn
#   Copyright (C) 2007 - 2009 Dan O'Reilly
#   Copyright (C) 2007 - 2009 Byron Hillis
#   Copyright (C) 2009        Andrew Psaltis
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License Version 2 as
#   published by the Free Software Foundation.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import random
import time
from string import maketrans, translate

import wpath
import misc
from misc import find_path 

# Regular expressions.
_re_mode = (re.I | re.M | re.S)
essid_pattern = re.compile('.*ESSID:"?(.*?)"?\s*\n', _re_mode)
ap_mac_pattern = re.compile('.*Address: (.*?)\n', _re_mode)
channel_pattern = re.compile('.*Channel:?=? ?(\d+)', _re_mode)
strength_pattern = re.compile('.*Quality:?=? ?(\d+)\s*/?\s*(\d*)', _re_mode)
altstrength_pattern = re.compile('.*Signal level:?=? ?(\d+)\s*/?\s*(\d*)', _re_mode)
signaldbm_pattern = re.compile('.*Signal level:?=? ?(-\d\d*)', _re_mode)
bitrates_pattern = re.compile('(\d+\s+\S+/s)', _re_mode)
mode_pattern = re.compile('.*Mode:([A-Za-z-]*?)\n', _re_mode)
freq_pattern = re.compile('.*Frequency:(.*?)\n', _re_mode)
wep_pattern = re.compile('.*Encryption key:(.*?)\n', _re_mode)
altwpa_pattern = re.compile('(wpa_ie)', _re_mode)
wpa1_pattern = re.compile('(WPA Version 1)', _re_mode)
wpa2_pattern = re.compile('(WPA2)', _re_mode)

#iwconfig-only regular expressions.
ip_up = re.compile(r'flags=[0.9]*<([^>]*)>', re.S)
ip_pattern = re.compile(r'inet [Aa]d?dr[^.]*:([^.]*\.[^.]*\.[^.]*\.[0-9]*)', re.S)
ip_pattern1 = re.compile(r'inet ([^.]*\.[^.]*\.[^.]*\.[0-9]*)', re.S)
bssid_pattern = re.compile('.*[(Access Point)|(Cell)]: (([0-9A-Z]{2}:){5}[0-9A-Z]{2})', _re_mode)
bitrate_pattern = re.compile('.*Bit Rate[=:](.*?/s)', _re_mode)
opmode_pattern = re.compile('.*Mode:(.*?) ', _re_mode)
authmethods_pattern = re.compile('.*Authentication capabilities :\n(.*?)Current', _re_mode)

# Regular expressions for wpa_cli output
auth_pattern = re.compile('.*wpa_state=(.*?)\n', _re_mode)

RALINK_DRIVER = 'ralink legacy'
NONE_DRIVER = 'none'

blacklist_strict = '!"#$%&\'()*+,./:;<=>?@[\\]^`{|}~ '
blacklist_norm = ";`$!*|><&\\"
blank_trans = maketrans("", "")

def _sanitize_string(string):
    if string:
        return translate(str(string), blank_trans, blacklist_norm)
    else:
        return string

def _sanitize_string_strict(string):
    if string:
        return translate(str(string), blank_trans, blacklist_strict)
    else:
        return string
  
_cache = {}
def timedcache(duration=5):
    """ A caching decorator for use with wnettools methods.
    
    Caches the results of a function for a given number of
    seconds (defaults to 5).
    
    """
    def _timedcache(f):
        def __timedcache(self, *args, **kwargs):
            key = str(args) + str(kwargs) + str(f)
            if hasattr(self, 'iface'):
                key += self.iface
            if (key in _cache and 
                (time.time() - _cache[key]['time']) < duration):
                return _cache[key]['value']
            else:
                value = f(self, *args, **kwargs)
                _cache[key] = { 'time' : time.time(), 'value' : value }
                return value
            
        return __timedcache
    
    return _timedcache

def GetDefaultGateway():
    """ Attempts to determine the default gateway by parsing route -n. """
    route_info = misc.Run("route -n")
    lines = route_info.split('\n')
    gateway = None
    for line in lines:
        words = line.split()
        print words
        if not words:
            continue
        if words[0] == '0.0.0.0':
            gateway = words[1]
            break
        
    if not gateway:
        print 'couldn\'t retrieve default gateway from route -n'
    return gateway

def GetWirelessInterfaces():
    """ Get available wireless interfaces.

    Attempts to get an interface first by parsing /proc/net/wireless,
    and should that fail, by parsing iwconfig.
    Returns:
    The first interface available.

    """
    dev_dir = '/sys/class/net/'
    ifnames = [iface for iface in os.listdir(dev_dir)
               if os.path.isdir(dev_dir + iface) and 
                  'wireless' in os.listdir(dev_dir + iface)]
    
    return ifnames

def GetWiredInterfaces():
    """ Returns a list of wired interfaces on the system. """
    basedir = '/sys/class/net/'
    return [iface for iface in os.listdir(basedir)
            if os.path.isdir(basedir + iface) and not 'wireless'
            in os.listdir(basedir + iface) and
            open(basedir + iface + "/type").readlines()[0].strip() == "1"]

def NeedsExternalCalls():
    """ Returns True if the backend needs to use an external program. """
    raise NotImplementedError

def GetWpaSupplicantDrivers():
    """ Returns a list of all valid wpa_supplicant drivers. """
    output = misc.Run(["wpa_supplicant", "-h"])
    try:
        output = output.split("drivers:")[1].split("options:")[0].strip()
    except:
        print "Warning: Couldn't get list of valid wpa_supplicant drivers"
        return [""]
    patt = re.compile("(\S+)\s+=.*")
    drivers = patt.findall(output) or [""]
    # We cannot use the "wired" driver for wireless interfaces.
    if 'wired' in drivers:
        drivers.remove('wired')
    return drivers
def IsValidWpaSuppDriver(driver):
    """ Returns True if given string is a valid wpa_supplicant driver. """
    output = misc.Run(["wpa_supplicant", "-D%s" % driver, "-iolan19",
                       "-c/etc/abcd%sdefzz.zconfz" % random.randint(1, 1000)])
    return not "Unsupported driver" in output
    
def neediface(default_response):
    """ A decorator for only running a method if self.iface is defined.
    
    This decorator is wrapped around Interface methods, and will
    return a provided default_response value if self.iface is not
    defined.
    
    """
    def wrapper(func):
        def newfunc(self, *args, **kwargs):
            if not self.iface or \
               not os.path.exists('/sys/class/net/%s' % self.iface):
                return default_response
            return func(self, *args, **kwargs)
        newfunc.__dict__ = func.__dict__
        newfunc.__doc__ = func.__doc__
        newfunc.__module__ = func.__module__
        return newfunc
    return wrapper


class BaseInterface(object):
    """ Control a network interface. """
    def __init__(self, iface, verbose=False):
        """ Initialise the object.

        Keyword arguments:
        iface -- the name of the interface
        verbose -- whether to print every command run

        """
        self.iface = _sanitize_string_strict(iface)
        self.verbose = verbose
        self.DHCP_CLIENT = None
        self.flush_tool = None
        self.link_detect = None       
        self.dhcp_object = None
    
    def SetDebugMode(self, value):
        """ If True, verbose output is enabled. """
        self.verbose = value

    def SetInterface(self, iface):
        """ Sets the interface.
        
        Keyword arguments:
        iface -- the name of the interface.
        
        """
        self.iface = _sanitize_string_strict(str(iface))
        
    def _find_program_path(self, program):
        """ Determines the full path for the given program.
        
        Searches for a given program name on the PATH.
        
        Keyword arguments:
        program -- The name of the program to search for
        
        Returns:
        The full path of the program or None
        
        """
        path = find_path(program)
        if not path and self.verbose:
            print "WARNING: No path found for %s" % program
        return path

    
    def _get_dhcp_command(self, flavor=None, hostname=None):
        """ Returns the correct DHCP client command. 
       
        Given a type of DHCP request (create or release a lease),
        this method will build a command to complete the request
        using the correct dhcp client, and cli options.
        
        """
        def get_client_name(cl):
            """ Converts the integer value for a dhcp client to a string. """
            if self.dhcpcd_cmd and cl in [misc.DHCPCD, misc.AUTO]:
                client = "dhcpcd"
                cmd = self.dhcpcd_cmd
            elif self.pump_cmd and cl in [misc.PUMP, misc.AUTO]: 
                client = "pump"
                cmd = self.pump_cmd
            elif self.dhclient_cmd and cl in [misc.DHCLIENT, misc.AUTO]:
                client = "dhclient"
                cmd = self.dhclient_cmd
                if self.dhclient_needs_verbose:
                    cmd += ' -v'
            elif self.udhcpc_cmd and cl in [misc.UDHCPC, misc.AUTO]:
                client = "udhcpc"
                cmd = self.udhcpc_cmd
            else:
                client = None
                cmd = ""
            return (client, cmd)

                # probably /var/lib/wicd/dhclient.conf with defaults
        dhclient_conf_path = os.path.join(
                    wpath.varlib,
                    'dhclient.conf'
                )
        
        client_dict = {
            "dhclient" : 
                {'connect' : r"%(cmd)s -cf %(dhclientconf)s %(iface)s",
                 'release' : r"%(cmd)s -r %(iface)s",
                 'id' : misc.DHCLIENT, 
                 },
            "pump" : 
                { 'connect' : r"%(cmd)s -i %(iface)s -h %(hostname)s",
                  'release' : r"%(cmd)s -r -i %(iface)s",
                  'id' : misc.PUMP,
                },
            "dhcpcd" : 
                {'connect' : r"%(cmd)s -h %(hostname)s --noipv4ll %(iface)s ",
                 'release' : r"%(cmd)s -k %(iface)s",
                 'id' : misc.DHCPCD,
                },
            "udhcpc":
                {'connect' : r"%(cmd)s -n -i %(iface)s -H %(hostname)s ",
                 'release' : r"killall -SIGUSR2 %(cmd)s",
                 'id' : misc.UDHCPC,
                },
        }
        (client_name, cmd) = get_client_name(self.DHCP_CLIENT)

        # cause dhclient doesn't have a handy dandy argument
        # for specifing the hostname to be sent
        if client_name == "dhclient" and flavor:
            if hostname == None:
                # <hostname> will use the system hostname
                # we'll use that if there is hostname passed
                # that shouldn't happen, though
                hostname = '<hostname>'
            print 'attempting to set hostname with dhclient'
            print 'using dhcpcd or another supported client may work better'
            dhclient_template = \
                open(os.path.join(wpath.etc, 'dhclient.conf.template'), 'r')

            output_conf = open(dhclient_conf_path, 'w')

            for line in dhclient_template.readlines():
                line = line.replace('$_HOSTNAME', hostname)
                output_conf.write(line)

            output_conf.close()
            dhclient_template.close()
            os.chmod(dhclient_conf_path, 0644)

        if not client_name or not cmd:
            print "WARNING: Failed to find a valid dhcp client!"
            return ""
            
        if flavor == "connect":
            if not hostname:
                hostname = os.uname()[1]
            return client_dict[client_name]['connect'] % \
                    { "cmd" : cmd,
                      "iface" : self.iface,
                      "hostname" : hostname,
                      'dhclientconf' : dhclient_conf_path }
        elif flavor == "release":
            return client_dict[client_name]['release'] % {"cmd":cmd, "iface":self.iface}
        else:
            return client_dict[client_name]['id']
    
    def AppAvailable(self, app):
        """ Return whether a given app is available.
        
        Given the name of an executable, determines if it is
        available for use by checking for a defined 'app'_cmd
        instance variable.
        
        """
        return bool(self.__dict__.get("%s_cmd" % app.replace("-", "")))
        
    def Check(self):
        """ Check that all required tools are available. """
        # THINGS TO CHECK FOR: ethtool, pptp-linux, dhclient, host
        self.CheckDHCP()
        self.CheckWiredTools()
        self.CheckWirelessTools()
        self.CheckSudoApplications()
        self.CheckRouteFlushTool()
        self.CheckResolvConf()

    def CheckResolvConf(self):
        """ Checks for the existence of resolvconf."""
        self.resolvconf_cmd = self._find_program_path("resolvconf")
        
    def CheckDHCP(self):
        """ Check for the existence of valid DHCP clients. 
        
        Checks for the existence of a supported DHCP client.  If one is
        found, the appropriate values for DHCP_CMD, DHCP_RELEASE, and
        DHCP_CLIENT are set.  If a supported client is not found, a
        warning is printed.
        
        """
        self.dhclient_cmd = self._find_program_path("dhclient")
        if self.dhclient_cmd != None:
            output = misc.Run(self.dhclient_cmd + " --version",
                    include_stderr=True)
            if '4.' in output:
                self.dhclient_needs_verbose = True
            else:
                self.dhclient_needs_verbose = False
        self.dhcpcd_cmd = self._find_program_path("dhcpcd")
        self.pump_cmd = self._find_program_path("pump")
        self.udhcpc_cmd = self._find_program_path("udhcpc")
        
    def CheckWiredTools(self):
        """ Check for the existence of ethtool and mii-tool. """
        self.miitool_cmd = self._find_program_path("mii-tool")
        self.ethtool_cmd = self._find_program_path("ethtool")
            
    def CheckWirelessTools(self):
        """ Check for the existence of wpa_cli """
        self.wpa_cli_cmd = self._find_program_path("wpa_cli")
        if not self.wpa_cli_cmd:
            print "wpa_cli not found.  Authentication will not be validated."
     
    def CheckRouteFlushTool(self):
        """ Check for a route flush tool. """
        self.ip_cmd = self._find_program_path("ip")
        self.route_cmd = self._find_program_path("route")
            
    def CheckSudoApplications(self):
        self.gksudo_cmd = self._find_program_path("gksudo")
        self.kdesu_cmd = self._find_program_path("kdesu")
        self.ktsuss_cmd = self._find_program_path("ktsuss")

    @neediface(False)
    def Up(self):
        """ Bring the network interface up.
        
        Returns:
        True
        
        """
        cmd = 'ifconfig ' + self.iface + ' up'
        if self.verbose: print cmd
        misc.Run(cmd)
        return True

    @neediface(False)
    def Down(self):
        """ Take down the network interface. 
        
        Returns:
        True
        
        """
        cmd = 'ifconfig ' + self.iface + ' down'
        if self.verbose: print cmd
        misc.Run(cmd)
        return True
    
    @timedcache(2)
    @neediface("")
    def GetIfconfig(self):
        """ Runs ifconfig and returns the output. """
        cmd = "ifconfig %s" % self.iface
        if self.verbose: print cmd
        return misc.Run(cmd)

    @neediface("")
    def SetAddress(self, ip=None, netmask=None, broadcast=None):
        """ Set the IP addresses of an interface.

        Keyword arguments:
        ip -- interface IP address in dotted quad form
        netmask -- netmask address in dotted quad form
        broadcast -- broadcast address in dotted quad form

        """
        for val in [ip, netmask, broadcast]:
            if not val:
                continue
            if not misc.IsValidIP(val):
                print 'WARNING: Invalid IP address found, aborting!'
                return False
        
        cmd = ''.join(['ifconfig ', self.iface, ' '])
        if ip:
            cmd = ''.join([cmd, ip, ' '])
        if netmask:
            cmd = ''.join([cmd, 'netmask ', netmask, ' '])
        if broadcast:
            cmd = ''.join([cmd, 'broadcast ', broadcast, ' '])
        if self.verbose: print cmd
        misc.Run(cmd)

    def _parse_dhclient(self, pipe):
        """ Parse the output of dhclient.
        
        Parses the output of dhclient and returns the status of
        the connection attempt.

        Keyword arguments:
        pipe -- stdout pipe to the dhclient process.
        
        Returns:
        'success' if succesful', an error code string otherwise.
        
        """
        dhclient_complete = False
        dhclient_success = False
        
        while not dhclient_complete:
            line = pipe.readline()
            if line == '':  # Empty string means dhclient is done.
                dhclient_complete = True
            else:
                print misc.to_unicode(line.strip('\n'))
            if line.startswith('bound'):
                dhclient_success = True
                dhclient_complete = True
                
        return self._check_dhcp_result(dhclient_success)
        
    def _parse_pump(self, pipe):
        """ Determines if obtaining an IP using pump succeeded.

        Keyword arguments:
        pipe -- stdout pipe to the pump process.
        
        Returns:
        'success' if succesful, an error code string otherwise.
        
        """
        pump_complete = False
        pump_success = True
        
        while not pump_complete:
            line = pipe.readline()
            if line == '':
                pump_complete = True
            elif line.strip().lower().startswith('Operation failed.'):
                pump_success = False
                pump_complete = True
            print misc.to_unicode(line)
    
        return self._check_dhcp_result(pump_success)

    def _parse_dhcpcd(self, pipe):
        """ Determines if obtaining an IP using dhcpcd succeeded.
        
        Keyword arguments:
        pipe -- stdout pipe to the dhcpcd process.
        
        Returns:
        'success' if succesful, an error code string otherwise.
        
        """
        dhcpcd_complete = False
        dhcpcd_success = True
        
        while not dhcpcd_complete:
            line = pipe.readline()
            if "Error" in line or "timed out" in line:
                dhcpcd_success = False
                dhcpcd_complete = True
            elif line == '':
                dhcpcd_complete = True
            print misc.to_unicode(line)
            
        return self._check_dhcp_result(dhcpcd_success)

    def _parse_udhcpc(self, pipe):
        """ Determines if obtaining an IP using udhcpc succeeded.

        Keyword arguments:
        pipe -- stdout pipe to the dhcpcd process.

        Returns:
        'success' if successful, an error code string otherwise.

        """
        udhcpc_complete = False
        udhcpc_success = True

        while not udhcpc_complete:
            line = pipe.readline()
            if line.endswith("failing."):
                udhcpc_success = False
                udhcpc_complete = True
            elif line == '':
                udhcpc_complete = True
            print line

        return self._check_dhcp_result(udhcpc_success)

    def _check_dhcp_result(self, success):
        """ Print and return the correct DHCP connection result. 
        
        Keyword Arguments:
        success -- boolean specifying if DHCP was succesful.
        
        Returns:
        'success' if success == True, 'dhcp_failed' otherwise.
        
        """
        if success:
            print 'DHCP connection successful'
            return 'success'
        else:
            print 'DHCP connection failed'
            return 'dhcp_failed'
            
    @neediface(False)
    def StartDHCP(self, hostname):
        """ Start the DHCP client to obtain an IP address.

        Keyword Arguments:
        hostname -- the hostname to send to the DHCP server
        
        Returns:
        A string representing the result of the DHCP command.  See
        _check_dhcp_result for the possible values.
        
        """
        cmd = self._get_dhcp_command('connect', hostname)
        if self.verbose: print cmd
        self.dhcp_object = misc.Run(cmd, include_stderr=True, return_obj=True)
        pipe = self.dhcp_object.stdout
        client_dict = { misc.DHCLIENT : self._parse_dhclient,
                        misc.DHCPCD : self._parse_dhcpcd,
                        misc.PUMP : self._parse_pump,
                        misc.UDHCPC : self._parse_udhcpc,
                      }
        
        DHCP_CLIENT = self._get_dhcp_command()
        if DHCP_CLIENT in client_dict:
            ret = client_dict[DHCP_CLIENT](pipe)
        else:
            print "ERROR: no dhcp client found"
            ret = None
        self.dhcp_object.wait()
        return ret
        
    @neediface(False)
    def ReleaseDHCP(self):
        """ Release the DHCP lease for this interface. """
        cmd = self._get_dhcp_command("release")
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface(False)
    def DelDefaultRoute(self):
        """ Delete only the default route for a device. """
        if self.ip_cmd and self.flush_tool in [misc.AUTO, misc.IP]:
            cmd = '%s route del default dev %s' % (self.ip_cmd, self.iface)
        elif self.route_cmd and self.flush_tool in [misc.AUTO, misc.ROUTE]:
            cmd = '%s del default dev %s' % (self.route_cmd, self.iface)
        else:
            print "No route manipulation command available!"
            return 
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface(False)
    def SetDNS(self, dns1=None, dns2=None, dns3=None, 
               dns_dom=None, search_dom=None):
        """ Set the DNS of the system to the specified DNS servers.

        Opens up resolv.conf and writes in the nameservers.

        Keyword arguments:
        dns1 -- IP address of DNS server 1
        dns2 -- IP address of DNS server 2
        dns3 -- IP address of DNS server 3
        dns_dom -- DNS domain
        search_dom -- DNS search domain

        """
        resolv_params = ""
        if dns_dom:
            resolv_params += 'domain %s\n' % dns_dom
        if search_dom:
            resolv_params += 'search %s\n' % search_dom

        valid_dns_list = []
        for dns in (dns1, dns2, dns3):
            if dns:
                if misc.IsValidIP(dns):
                    if self.verbose:
                        print 'Setting DNS : ' + dns
                    valid_dns_list.append("nameserver %s\n" % dns)
                else:
                    print 'DNS IP %s is not a valid IP address, skipping' % dns

        if valid_dns_list:
            resolv_params += ''.join(valid_dns_list)

        if self.resolvconf_cmd:
            cmd = [self.resolvconf_cmd, '-a', self.iface]
            if self.verbose: print cmd
            p = misc.Run(cmd, include_stderr=True, return_obj=True)
            p.communicate(input=resolv_params)
        else:
            resolv = open("/etc/resolv.conf", "w")
            resolv.write(resolv_params + "\n")
            resolv.close()
        
    @neediface(False)
    def FlushRoutes(self):
        """ Flush network routes for this device. """
        if self.ip_cmd and self.flush_tool in [misc.AUTO, misc.IP]:
            cmds = ['%s route flush dev %s' % (self.ip_cmd, self.iface)]
        elif self.route_cmd and self.flush_tool in [misc.AUTO, misc.ROUTE]:
            cmds = ['%s del dev %s' % (self.route_cmd, self.iface)]
        else:
            print "No flush command available!"
            cmds = []
        for cmd in cmds:
            if self.verbose: print cmd
            misc.Run(cmd)

    @neediface(False)
    def SetDefaultRoute(self, gw):
        """ Add a default route with the specified gateway.

        Keyword arguments:
        gw -- gateway of the default route in dotted quad form

        """
        if not misc.IsValidIP(gw):
            print 'WARNING: Invalid gateway found.  Aborting!'
            return False
        cmd = 'route add default gw %s dev %s' % (gw, self.iface)
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface("")
    def GetIP(self, ifconfig=""):
        """ Get the IP address of the interface.

        Returns:
        The IP address of the interface in dotted quad form.

        """
        if not ifconfig:
            output = self.GetIfconfig()
        else:
            output = ifconfig
        # check multiple ifconfig output styles
        for pat in [ip_pattern, ip_pattern1]:
            m = misc.RunRegex(pat, output)
            if m: return m
        return None

    @neediface(False)
    def VerifyAPAssociation(self, gateway):
        """ Verify assocation with an access point. 
        
        Verifies that an access point can be contacted by
        trying to ping it.
        
        """
        if "iputils" in misc.Run(["ping", "-V"]):
            cmd = "ping -q -w 3 -c 1 %s" % gateway
        else:
            # ping is from inetutils-ping (which doesn't support -w)
            # or from some other package
            #
            # If there's no AP association, this will wait for
            # timeout, while the above will wait (-w) 3 seconds at
            # most.
            cmd = "ping -q -c 1 %s" % gateway
        if self.verbose: print cmd
        return misc.LaunchAndWait(cmd)

    @neediface(False)
    def IsUp(self, ifconfig=None):
        """ Determines if the interface is up.

        Returns:
        True if the interface is up, False otherwise.

        """
        flags_file = '/sys/class/net/%s/flags' % self.iface
        try:
            flags = open(flags_file, "r").read().strip()
        except IOError:
            print "Could not open %s, using ifconfig to determine status" % flags_file
            return self._slow_is_up(ifconfig)
        return bool(int(flags, 16) & 1)
    
    @neediface(False)
    def StopWPA(self):
        """ Terminates wpa using wpa_cli"""
        cmd = 'wpa_cli -i %s terminate' % self.iface
        if self.verbose: print cmd
        misc.Run(cmd)
        
        
    def _slow_is_up(self, ifconfig=None):
        """ Determine if an interface is up using ifconfig. """
        if not ifconfig:
            output = self.GetIfconfig()
        else:
            output = ifconfig
        lines = output.split('\n')
        if len(lines) < 5:
            return False
        # check alternative ifconfig output style
        m = misc.RunRegex(ip_up, lines[0])
        if m and m.find('UP') > -1:
            return True
        # check classic ifconfig output style
        for line in lines[1:4]:
            if line.strip().startswith('UP'):
                return True   
        return False


class BaseWiredInterface(BaseInterface):
    """ Control a wired network interface. """
    def __init__(self, iface, verbose=False):
        """ Initialise the wired network interface class.

        Keyword arguments:
        iface -- name of the interface
        verbose -- print all commands

        """
        BaseInterface.__init__(self, iface, verbose)

    @neediface(False)
    def GetPluggedIn(self):
        """ Get the current physical connection state.

        The method will first attempt to use ethtool do determine
        physical connection state.  Should ethtool fail to run properly,
        mii-tool will be used instead.

        Returns:
        True if a link is detected, False otherwise.

        """
        # check for link using /sys/class/net/iface/carrier
        # is usually more accurate
        sys_device = '/sys/class/net/%s/' % self.iface
        carrier_path = sys_device + 'carrier'
        if not self.IsUp():
            MAX_TRIES = 3
            tries = 0
            self.Up()
            while True:
                tries += 1
                time.sleep(2)
                if self.IsUp() or tries > MAX_TRIES: break
      
        if os.path.exists(carrier_path):
            carrier = open(carrier_path, 'r')
            try:
                link = carrier.read().strip()
                link = int(link)
                if link == 1:
                    return True
                elif link == 0:
                    return False
            except (IOError, ValueError, TypeError):
                print 'Error checking link using /sys/class/net/%s/carrier' % self.iface
                
        if self.ethtool_cmd and self.link_detect in [misc.ETHTOOL, misc.AUTO]:
            return self._eth_get_plugged_in()
        elif self.miitool_cmd and self.link_detect in [misc.MIITOOL, misc.AUTO]:
            return self._mii_get_plugged_in()
        else:
            print ('Error: No way of checking for a wired connection. Make ' +
                   'sure that either mii-tool or ethtool is installed.')
            return False

    def _eth_get_plugged_in(self):
        """ Use ethtool to determine the physical connection state.
        
        Returns:
        True if a link is detected, False otherwise.
        
        """
        cmd = "%s %s" % (self.ethtool_cmd, self.iface)
        if not self.IsUp():
            print 'Wired Interface is down, putting it up'
            self.Up()
            time.sleep(6)
        if self.verbose: print cmd
        tool_data = misc.Run(cmd, include_stderr=True)
        if misc.RunRegex(re.compile('(Link detected: yes)', re.I | re.M  | re.S),
                         tool_data):
            return True
        else:
            return False
    
    def _mii_get_plugged_in(self):
        """ Use mii-tool to determine the physical connection state. 
                
        Returns:
        True if a link is detected, False otherwise.
        
        """
        cmd = "%s %s" % (self.miitool_cmd, self.iface)
        if self.verbose: print cmd
        tool_data = misc.Run(cmd, include_stderr=True)
        if misc.RunRegex(re.compile('(Invalid argument)', re.I | re.M  | re.S), 
                         tool_data) is not None:
            print 'Wired Interface is down, putting it up'
            self.Up()
            time.sleep(4)
            if self.verbose: print cmd
            tool_data = misc.Run(cmd, include_stderr=True)
        
        if misc.RunRegex(re.compile('(link ok)', re.I | re.M | re.S),
                         tool_data) is not None:
            return True
        else:
            return False
        
    def Authenticate(self, network):
        misc.ParseEncryption(network)
        cmd = ['wpa_supplicant', '-B', '-i', self.iface, '-c',
               os.path.join(wpath.networks, 'wired'),
               '-Dwired']
        if self.verbose: print cmd
        misc.Run(cmd)

class BaseWirelessInterface(BaseInterface):
    """ Control a wireless network interface. """
    def __init__(self, iface, verbose=False, wpa_driver='wext'):
        """ Initialise the wireless network interface class.

        Keyword arguments:
        iface -- name of the interface
        verbose -- print all commands

        """
        BaseInterface.__init__(self, iface, verbose)
        self.wpa_driver = wpa_driver
        self.scan_iface = None
        
    def SetWpaDriver(self, driver):
        """ Sets the wpa_driver. """
        self.wpa_driver = _sanitize_string(driver)

    @neediface(False)
    def SetEssid(self, essid):
        """ Set the essid of the wireless interface.

        Keyword arguments:
        essid -- essid to set the interface to

        """
        cmd = ['iwconfig', self.iface, 'essid', '--', str(essid)]
        if self.verbose: print str(cmd)
        misc.Run(cmd)

    @neediface(False)
    def GetKillSwitchStatus(self):
        """ Determines if the wireless killswitch is enabled.
        
        Returns:
        True if the killswitch is enabled, False otherwise.
        
        """
        output = self.GetIwconfig()

        killswitch_pattern = re.compile('.*radio off', re.I | re.M | re.S)
        if killswitch_pattern.search(output):
            radiostatus = True
        else:
            radiostatus = False

        return radiostatus
    
    @timedcache(2)
    @neediface(False)
    def GetIwconfig(self):
        """ Returns the output of iwconfig for this interface. """
        cmd = "iwconfig " + self.iface
        if self.verbose: print cmd
        return misc.Run(cmd)

    def _FreqToChannel(self, freq):
        """ Translate the specified frequency to a channel.

        Note: This function is simply a lookup dict and therefore the
        freq argument must be in the dict to provide a valid channel.

        Keyword arguments:
        freq -- string containing the specified frequency

        Returns:
        The channel number, or None if not found.

        """
        ret = None
        freq_dict = {'2.412 GHz': 1, '2.417 GHz': 2, '2.422 GHz': 3,
                         '2.427 GHz': 4, '2.432 GHz': 5, '2.437 GHz': 6,
                         '2.442 GHz': 7, '2.447 GHz': 8, '2.452 GHz': 9,
                         '2.457 GHz': 10, '2.462 GHz': 11, '2.467 GHz': 12,
                         '2.472 GHz': 13, '2.484 GHz': 14 }
        try:
            ret = freq_dict[freq]
        except KeyError:
            print "Couldn't determine channel number for frequency: " + str(freq)
        
        return ret

    def _GetRalinkInfo(self):
        """ Get a network info list used for ralink drivers
    
        Calls iwpriv <wireless interface> get_site_survey, which
        on some ralink cards will return encryption and signal
        strength info for wireless networks in the area.
    
        """
        iwpriv = misc.Run('iwpriv ' + self.iface + ' get_site_survey')
        if self.verbose:
            print iwpriv
        lines = iwpriv.splitlines()[2:]
        aps = {}
        patt = re.compile("((?:[0-9A-Z]{2}:){5}[0-9A-Z]{2})")
        for x in lines:
            ap = {}
            info = x.split("   ")
            info = filter(None, [x.strip() for x in info])
            if len(info) < 5:
                continue
            if re.match(patt, info[2].upper()):
                bssid = info[2].upper()
                offset = -1
            elif re.match(patt, info[3].upper()):
                bssid = info[3].upper()
                offset = 0
            else:  # Invalid
                print 'Invalid iwpriv line.  Skipping it.'
                continue
            ap['nettype'] = info[-1]
            ap['strength'] = info[1]
            if info[4 + offset] == 'WEP':
                ap['encryption_method'] = 'WEP'
                ap['enctype'] = 'WEP'
                ap['keyname'] = 'Key1'
                ap['authmode'] = info[5 + offset]
            elif info[5 + offset] in ['WPA-PSK', 'WPA']:
                ap['encryption_method'] = 'WPA'
                ap['authmode'] = "WPAPSK"
                ap['keyname'] = "WPAPSK"
                ap['enctype'] = info[4 + offset]
            elif info[5 + offset] == 'WPA2-PSK':
                ap['encryption_method'] = 'WPA2'
                ap['authmode'] ="WPA2PSK"
                ap['keyname'] = "WPA2PSK"
                ap['enctype'] = info[4 + offset]
            elif info[4 + offset] == "NONE":
                ap['encryption_method'] = None
            else:
                print "Unknown AuthMode, can't assign encryption_method!"
                ap['encryption_method'] = 'Unknown'
            aps[bssid] = ap
        if self.verbose: print str(aps)
        return aps

    def _ParseRalinkAccessPoint(self, ap, ralink_info, cell):
        """ Parse encryption and signal strength info for ralink cards

        Keyword arguments:
        ap -- array containing info about the current access point
        ralink_info -- dict containing available network info
        cell -- string containing cell information

        Returns:
        Updated array containing info about the current access point

        """
        wep_pattern = re.compile('.*Encryption key:(.*?)\n', re.I | re.M  | re.S)
        if ralink_info.has_key(ap['bssid']):
            info = ralink_info[ap['bssid']]
            for key in info.keys():
                ap[key] = info[key]
            if misc.RunRegex(wep_pattern, cell) == 'on':
                ap['encryption'] = True
            else:
                ap['encryption'] = False
        return ap

    @neediface(False)
    def SetMode(self, mode):
        """ Set the mode of the wireless interface.

        Keyword arguments:
        mode -- mode to set the interface to

        """
        mode = _sanitize_string_strict(mode)
        if not mode:
            mode = 'master'
        if mode.lower() == 'master':
            mode = 'managed'
        cmd = 'iwconfig %s mode %s' % (self.iface, mode)
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface(False)
    def SetChannel(self, channel):
        """ Set the channel of the wireless interface.

        Keyword arguments:
        channel -- channel to set the interface to

        """
        if not channel.isdigit():
            print 'WARNING: Invalid channel found.  Aborting!'
            return False
        
        cmd = 'iwconfig %s channel %s' % (self.iface, str(channel))
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface(False)
    def SetKey(self, key):
        """ Set the encryption key of the wireless interface.

        Keyword arguments:
        key -- encryption key to set

        """
        cmd = 'iwconfig %s key %s' % (self.iface, key)
        if self.verbose: print cmd
        misc.Run(cmd)

    @neediface(False)
    def Associate(self, essid, channel=None, bssid=None):
        """ Associate with the specified wireless network.

        Keyword arguments:
        essid -- essid of the network
        channel -- channel of the network
        bssid -- bssid of the network

        """
        self.SetEssid(essid)
        base = "iwconfig %s" % self.iface
        if channel and str(channel).isdigit():
            cmd = "%s channel %s" % (base, str(channel))
            if self.verbose: print cmd
            misc.Run(cmd)
        if bssid:
            cmd = "%s ap %s" % (base, bssid)
            if self.verbose: print cmd
            misc.Run(cmd)
        
    def GeneratePSK(self, network):
        """ Generate a PSK using wpa_passphrase. 

        Keyword arguments:
        network -- dictionary containing network info
        
        """
        wpa_pass_path = misc.find_path('wpa_passphrase')
        if not wpa_pass_path: return None
        key_pattern = re.compile('network={.*?\spsk=(.*?)\n}.*',
                                 re.I | re.M  | re.S)
        cmd = [wpa_pass_path, str(network['essid']), str(network['key'])]
        if self.verbose: print cmd
        return misc.RunRegex(key_pattern, misc.Run(cmd))

    @neediface(False)
    def Authenticate(self, network):
        """ Authenticate with the specified wireless network.

        Keyword arguments:
        network -- dictionary containing network info

        """
        misc.ParseEncryption(network)
        if self.wpa_driver == RALINK_DRIVER:
            self._AuthenticateRalinkLegacy(network)
        else:
            if self.wpa_driver == NONE_DRIVER:
                driver = ''
            else:
                driver = '-D' + self.wpa_driver
            cmd = ['wpa_supplicant', '-B', '-i', self.iface, '-c',
                   os.path.join(wpath.networks, 
                                network['bssid'].replace(':', '').lower()),
                   driver]
            if self.verbose: print cmd
            misc.Run(cmd)

    def _AuthenticateRalinkLegacy(self, network):
        """ Authenticate with the specified wireless network.

        This function handles Ralink legacy cards that cannot use
        wpa_supplicant.

        Keyword arguments:
        network -- dictionary containing network info

        """
        if network.get('key') != None:
            try:
                info = self._GetRalinkInfo()[network.get('bssid')]
            except KeyError:
                print "Could not find current network in iwpriv " + \
                      "get_site_survey results.  Cannot authenticate."
                return
            
            if info['enctype'] == "WEP" and info['authtype'] == 'OPEN':
                print 'Setting up WEP'
                cmd = ''.join(['iwconfig ', self.iface, ' key ',
                              network.get('key')])
                if self.verbose: print cmd
                misc.Run(cmd)
            else:
                cmd_list = []
                cmd_list.append('NetworkType=' + info['nettype'])
                cmd_list.append('AuthMode=' + info['authmode'])
                cmd_list.append('EncrypType=' + info['enctype'])
                cmd_list.append('SSID="%s"' % network['essid'])
                cmd_list.append('%s="%s"' % (network['keyname'], network['key']))
                if info['nettype'] == 'SHARED' and info['enctype'] == 'WEP':
                    cmd_list.append('DefaultKeyID=1')
    
                for cmd in cmd_list:
                    cmd = ['iwpriv', self.iface, 'set', cmd]
                    if self.verbose: print ' '.join(cmd)
                    misc.Run(cmd)

    @neediface([])
    def GetNetworks(self):
        """ Get a list of available wireless networks.

        Returns:
        A list containing available wireless networks.

        """
        cmd = 'iwlist ' + self.iface + ' scan'
        if self.verbose: print cmd
        results = misc.Run(cmd)
        # Split the networks apart, using Cell as our split point
        # this way we can look at only one network at a time.
        # The spaces around '   Cell ' are to minimize the chance that someone
        # has an essid named Cell...
        networks = results.split( '   Cell ' )

        # Get available network info from iwpriv get_site_survey
        # if we're using a ralink card (needed to get encryption info)
        if self.wpa_driver == RALINK_DRIVER:
            ralink_info = self._GetRalinkInfo()
        else:
            ralink_info = None

        # An array for the access points
        access_points = []
        access_points = {}
        for cell in networks:
            # Only use sections where there is an ESSID.
            if 'ESSID:' in cell:
                # Add this network to the list of networks
                entry = self._ParseAccessPoint(cell, ralink_info)
                if entry is not None:
                    # Normally we only get duplicate bssids with hidden
                    # networks.  If we hit this, we only want the entry
                    # with the real essid to be in the network list.
                    if (entry['bssid'] not in access_points 
                        or not entry['hidden']):
                        access_points[entry['bssid']] = entry

        return access_points.values()
    
    def _ParseAccessPoint(self, cell, ralink_info):
        """ Parse a single cell from the output of iwlist.

        Keyword arguments:
        cell -- string containing the cell information
        ralink_info -- string contating network information needed
                       for ralink cards.

        Returns:
        A dictionary containing the cell networks properties.

        """
        ap = {}
        ap['essid'] = misc.RunRegex(essid_pattern, cell)
        try:
            ap['essid'] = misc.to_unicode(ap['essid'])
        except (UnicodeDecodeError, UnicodeEncodeError):
            print 'Unicode problem with current network essid, ignoring!!'
            return None

        # We (well, DBus) don't support ESSIDs with null bytes in it.
        # From some bugreports, it seems like some APs transmit the "hidden"
        # essid as NULL bytes. Let's strip them off.
        ap['essid'] = ap['essid'].replace('\x00', '')

        if ap['essid'] in ['Hidden', '<hidden>', "", None]:
            print 'hidden'
            ap['hidden'] = True
            ap['essid'] = "<hidden>"
        else:
            ap['hidden'] = False

        # Channel - For cards that don't have a channel number,
        # convert the frequency.
        ap['channel'] = misc.RunRegex(channel_pattern, cell)
        if ap['channel'] == None:
            freq = misc.RunRegex(freq_pattern, cell)
            ap['channel'] = self._FreqToChannel(freq)

        # Bit Rate
        ap['bitrates'] = misc.RunRegex(bitrates_pattern,
                                       cell.split("Bit Rates")[-1])
     
        # BSSID
        ap['bssid'] = misc.RunRegex(ap_mac_pattern, cell)

        # Mode
        ap['mode'] = misc.RunRegex(mode_pattern, cell)
        if ap['mode'] is None:
            print 'Invalid network mode string, ignoring!'
            return None

        # Break off here if we're using a ralink card
        if self.wpa_driver == RALINK_DRIVER:
            ap = self._ParseRalinkAccessPoint(ap, ralink_info, cell)
        elif misc.RunRegex(wep_pattern, cell) == 'on':
            # Encryption - Default to WEP
            ap['encryption'] = True
            ap['encryption_method'] = 'WEP'

            if misc.RunRegex(wpa1_pattern, cell) == 'WPA Version 1':
                ap['encryption_method'] = 'WPA'

            if misc.RunRegex(altwpa_pattern, cell) == 'wpa_ie':
                ap['encryption_method'] = 'WPA'

            if misc.RunRegex(wpa2_pattern, cell) == 'WPA2':
                ap['encryption_method'] = 'WPA2'
        else:
            ap['encryption'] = False

        # Link Quality
        # Set strength to -1 if the quality is not found
        ap['quality'] = self._get_link_quality(cell)
        if ap['quality'] is None:
            ap['quality'] = -1

        # Signal Strength (only used if user doesn't want link
        # quality displayed or it isn't found)
        if misc.RunRegex(signaldbm_pattern, cell):
            ap['strength'] = misc.RunRegex(signaldbm_pattern, cell)
        elif self.wpa_driver != RALINK_DRIVER:  # This is already set for ralink
            ap['strength'] = -1

        return ap

    def ValidateAuthentication(self, auth_time):
        """ Validate WPA authentication.

            Validate that the wpa_supplicant authentication
            process was successful.

            NOTE: It's possible this could return False,
            though in reality wpa_supplicant just isn't
            finished yet.
            
            Keyword arguments:
            auth_time -- The time at which authentication began.
            
            Returns:
            True if wpa_supplicant authenticated succesfully,
            False otherwise.

        """
        # Right now there's no way to do this for these drivers
        if self.wpa_driver == RALINK_DRIVER or not self.wpa_cli_cmd:
            return True

        MAX_TIME = 35
        MAX_DISCONNECTED_TIME = 3
        disconnected_time = 0
        forced_rescan = False
        while (time.time() - auth_time) < MAX_TIME:
            cmd = '%s -i %s status' % (self.wpa_cli_cmd, self.iface)
            output = misc.Run(cmd)
            result = misc.RunRegex(auth_pattern, output)
            if self.verbose:
                print 'WPA_CLI RESULT IS', result

            if not result:
                return False
            if result == "COMPLETED":
                return True
            elif result == "DISCONNECTED" and not forced_rescan:
                disconnected_time += 1
                if disconnected_time > MAX_DISCONNECTED_TIME:
                    disconnected_time = 0
                    # Force a rescan to get wpa_supplicant moving again.
                    forced_rescan = True
                    self._ForceSupplicantScan()
                    MAX_TIME += 5
            else:
                disconnected_time = 0
            time.sleep(1)

        print 'wpa_supplicant authentication may have failed.'
        return False
        

    def _ForceSupplicantScan(self):
        """ Force wpa_supplicant to rescan available networks.
    
        This function forces wpa_supplicant to rescan.
        This works around authentication validation sometimes failing for
        wpa_supplicant because it remains in a DISCONNECTED state for 
        quite a while, after which a rescan is required, and then
        attempting to authenticate.  This whole process takes a long
        time, so we manually speed it up if we see it happening.
        
        """
        print 'wpa_supplicant rescan forced...'
        cmd = 'wpa_cli -i' + self.iface + ' scan'
        misc.Run(cmd)

    @neediface("")
    def GetBSSID(self, iwconfig=None):
        """ Get the MAC address for the interface. """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
            
        bssid = misc.RunRegex(bssid_pattern, output)
        return bssid

    @neediface("")
    def GetCurrentBitrate(self, iwconfig=None):
        """ Get the current bitrate for the interface. """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
            
        bitrate = misc.RunRegex(bitrate_pattern, output)
        return bitrate

    @neediface("")
    def GetOperationalMode(self, iwconfig=None):
        """ Get the operational mode for the interface. """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
            
        opmode = misc.RunRegex(opmode_pattern, output)
        if opmode:
            opmode = opmode.strip()
        return opmode

    @neediface("")
    def GetAvailableAuthMethods(self, iwlistauth=None):
        """ Get the available authentication methods for the interface. """
        if not iwlistauth:
            cmd = 'iwlist ' + self.iface + ' auth'
            if self.verbose: print cmd
            output = misc.Run(cmd)
        else:
            output = iwlistauth
            
        authm = misc.RunRegex(authmethods_pattern, output)
        authm_list = [m.strip() for m in authm.split('\n') if m.strip()]
        return ';'.join(authm_list)

    def _get_link_quality(self, output):
        """ Parse out the link quality from iwlist scan or iwconfig output. """
        try:
            [(strength, max_strength)] = strength_pattern.findall(output)
        except ValueError:
            (strength, max_strength) = (None, None)

        if strength in ['', None]:
            try:
                [(strength, max_strength)] = altstrength_pattern.findall(output)
            except ValueError:
                # if the pattern was unable to match anything
                # we'll return 101, which will allow us to stay
                # connected even though we don't know the strength
                # it also allows us to tell if 
                return 101
        if strength not in ['', None] and max_strength:
            return (100 * int(strength) // int(max_strength))
        elif strength not in ["", None]:
            return int(strength)
        else:
            return None

    @neediface(-1)
    def GetSignalStrength(self, iwconfig=None):
        """ Get the signal strength of the current network.

        Returns:
        The signal strength.

        """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
        return self._get_link_quality(output)
    
    @neediface(-100)
    def GetDBMStrength(self, iwconfig=None):
        """ Get the dBm signal strength of the current network.

        Returns:
        The dBm signal strength.

        """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
        dbm_strength = misc.RunRegex(signaldbm_pattern, output)
        return dbm_strength

    @neediface("")
    def GetCurrentNetwork(self, iwconfig=None):
        """ Get the essid of the current network.

        Returns:
        The current network essid.

        """
        if not iwconfig:
            output = self.GetIwconfig()
        else:
            output = iwconfig
        network = misc.to_unicode(misc.RunRegex(essid_pattern, output))
        if network:
            network = misc.to_unicode(network)
        return network
